"""Tabular iterators."""

from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import cast

import numpy as np

from lightautoml.reader.utils import set_sklearn_folds
from lightautoml.tasks import Task

from ..dataset.np_pd_dataset import CSRSparseDataset
from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from .base import CustomIdxs
from .base import CustomIterator
from .base import DummyIterator
from .base import HoldoutIterator
from .base import TrainValidIterator


NumpyOrSparse = Union[CSRSparseDataset, NumpyDataset, PandasDataset]


class FoldsIterator(TrainValidIterator):
    """Classic cv iterator.

    Folds should be defined in Reader, based on cross validation method.
    """

    def __init__(self, train: NumpyOrSparse, n_folds: Optional[int] = None):
        """Creates iterator.

        Args:
            train: Dataset for folding.
            n_folds: Number of folds.

        """
        assert hasattr(train, "folds"), "Folds in dataset should be defined to make folds iterator."

        self.train = train
        self.n_folds = train.folds.max() + 1
        if n_folds is not None:
            self.n_folds = min(self.n_folds, n_folds)

    def __len__(self) -> int:
        """Get len of iterator.

        Returns:
            Number of folds.

        """
        return self.n_folds

    def __iter__(self) -> "FoldsIterator":
        """Set counter to 0 and return self.

        Returns:
            Iterator for folds.

        """
        self._curr_idx = 0
        return self

    def __next__(self) -> Tuple[np.ndarray, NumpyOrSparse, NumpyOrSparse]:
        """Define how to get next object.

        Returns:
            Mask for current fold, train dataset, validation dataset.

        """
        if self._curr_idx == self.n_folds:
            raise StopIteration
        val_idx = self.train.folds == self._curr_idx
        tr_idx = np.logical_not(val_idx)
        idx = np.arange(self.train.shape[0])
        tr_idx, val_idx = idx[tr_idx], idx[val_idx]
        train, valid = self.train[tr_idx], self.train[val_idx]
        self._curr_idx += 1
        return val_idx, cast(NumpyOrSparse, train), cast(NumpyOrSparse, valid)

    def get_validation_data(self) -> NumpyOrSparse:
        """Just return train dataset.

        Returns:
            Whole train dataset.

        """
        return self.train

    def convert_to_holdout_iterator(self) -> HoldoutIterator:
        """Convert iterator to hold-out-iterator.

        Fold 0 is used for validation, everything else is used for training.

        Returns:
            new hold-out-iterator.

        """
        val_idx = self.train.folds == 0
        tr_idx = np.logical_not(val_idx)
        idx = np.arange(self.train.shape[0])
        tr_idx, val_idx = idx[tr_idx], idx[val_idx]
        train, valid = self.train[tr_idx], self.train[val_idx]
        return HoldoutIterator(train, valid)


def get_numpy_iterator(
    train: NumpyOrSparse,
    valid: Optional[NumpyOrSparse] = None,
    n_folds: Optional[int] = None,
    iterator: Optional[CustomIdxs] = None,
) -> Union[FoldsIterator, HoldoutIterator, CustomIterator, DummyIterator]:
    """Get iterator for np/sparse dataset.

    If valid is defined, other parameters are ignored.
    Else if iterator is defined n_folds is ignored.

    Else if n_folds is defined iterator will be created by folds index.
    Else ``DummyIterator`` - (train, train) will be created.

    Args:
        train: ``LAMLDataset`` to train.
        valid: Optional ``LAMLDataset`` for validate.
        n_folds: maximum number of folds to iterate. If ``None`` - iterate through all folds.
        iterator: Takes dataset as input and return an iterator of indexes of train/valid for train dataset.

    Returns:
        new train-validation iterator.

    """
    if valid is not None:
        train_valid = HoldoutIterator(train, valid)
    elif iterator is not None:
        train_valid = CustomIterator(train, iterator)
    elif train.folds is not None:
        train_valid = FoldsIterator(train, n_folds)
    else:
        train_valid = DummyIterator(train)

    return train_valid


class TimeSeriesIterator:
    """Time Series Iterator."""

    @staticmethod
    def split_by_dates(datetime_col, splitter):
        """Create indexes of folds splitted by thresholds.

        Args:
            datetime_col: Column with value which can be interpreted as time/ordinal value (ex: np.datetime64).
            splitter: List of thresholds (same value as ).

        Returns:
            folds: Array of folds' indexes.

        """
        splitter = np.sort(splitter)
        folds = np.searchsorted(splitter, datetime_col)

        return folds

    @staticmethod
    def split_by_parts(datetime_col, n_splits: int):
        """Create indexes of folds splitted into equal parts.

        Args:
            datetime_col: Column with value which can be interpreted as time/ordinal value (ex: np.datetime64).
            n_splits: Number of splits(folds).

        Returns:
            folds: Array of folds' indexes.

        """
        idx = np.arange(datetime_col.shape[0])
        order = np.argsort(datetime_col)
        sorted_idx = idx[order]
        folds = np.concatenate(
            [[n] * x.shape[0] for (n, x) in enumerate(np.array_split(sorted_idx, n_splits))],
            axis=0,
        )
        folds = folds[sorted_idx]

        return folds

    def __init__(
        self,
        datetime_col,
        n_splits: Optional[int] = 5,
        date_splits: Optional[Sequence] = None,
        sorted_kfold: bool = False,
    ):
        """Generates time series data split. Sorter - include left, exclude right.

        Args:
            datetime_col: Column with value which can be interpreted as time/ordinal value (ex: np.datetime64).
            n_splits: Number of splits.
            date_splits: List of thresholds.
            sorted_kfold: is sorted.

        """
        self.sorted_kfold = sorted_kfold

        if date_splits is not None:
            folds = self.split_by_dates(datetime_col, date_splits)
        elif n_splits is not None:
            folds = self.split_by_parts(datetime_col, n_splits)

        uniques = np.unique(folds)
        assert (uniques == np.arange(uniques.shape[0])).all(), "Fold splits is incorrect"
        # sort in descending order - for holdout from custom be the biggest part
        self.folds = uniques[::-1][folds]
        self.n_splits = uniques.shape[0]

    def __len__(self) -> int:
        """Get number of folds.

        Returns:
            length.

        """
        if self.sorted_kfold:
            return self.n_splits
        return self.n_splits - 1

    def __getitem__(self, item) -> Tuple[np.ndarray, np.ndarray]:
        """Select train/validation indexes.

        For Train indexes use all dates before Validation dates.

        Args:
            item: Index of fold.

        Returns:
            Tuple of train/validation indexes.

        """
        if item >= len(self):
            raise StopIteration

        idx = np.arange(self.folds.shape[0])

        if self.sorted_kfold:
            return idx[self.folds != item], idx[self.folds == item]

        return idx[self.folds < (item + 1)], idx[self.folds == (item + 1)]


class UpliftIterator:
    """Iterator for uplift modeling task.

    Generates time series data split. Sorter - include left, exclude right.

    Args:
        treatment_col: Treatment column: 0 - control group, 1 - treatment group.
        target: Target values.
        mode: Flag.
        task: Task.
        n_folds: Number of folds.

    """

    def __init__(
        self,
        treatment_col: np.ndarray,
        target: np.ndarray,
        mode: bool,
        task: Task,
        n_folds: int = 5,
    ):
        self.task = task
        self.n_folds = n_folds
        self.mode = mode

        idx = np.arange(treatment_col.shape[0])
        flg = treatment_col.astype(np.bool) == self.mode

        self.constant_idx = idx[flg]
        self.splitted_idx = idx[~flg]

        self.folds = set_sklearn_folds(self.task, target[self.splitted_idx], self.n_folds)

    def __len__(self):
        """Get number of folds.

        Returns:
            length.

        """
        return self.n_folds

    def __getitem__(self, item):
        """Select train/validation indexes.

        For Train indexes use all dates before Validation dates.

        Args:
            item: index of fold.

        Returns:
            Tuple of train/validation indexes.

        """
        if item + 1 >= self.__len__():
            raise IndexError()

        val_idx = self.splitted_idx[self.folds == item]
        train_fold_idx = self.splitted_idx[self.folds != item]

        train_idx = np.concatenate([self.constant_idx, train_fold_idx])

        return train_idx, val_idx



"""Base classes for machine learning algorithms."""

import logging

from abc import ABC
from abc import abstractmethod
from copy import copy
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import cast

import numpy as np

from lightautoml.validation.base import TrainValidIterator

from ..dataset.base import LAMLDataset
from ..dataset.np_pd_dataset import CSRSparseDataset
from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.np_pd_dataset import PandasDataset
from ..dataset.roles import NumericRole
from ..utils.timer import PipelineTimer
from ..utils.timer import TaskTimer


logger = logging.getLogger(__name__)
TabularDataset = Union[NumpyDataset, CSRSparseDataset, PandasDataset]


class MLAlgo(ABC):
    """Abstract class for machine learning algorithm.

    Assume that features are already selected,
    but parameters my be tuned and set before training.

    Args:
        default_params: Algo hyperparams.
        freeze_defaults:
            - ``True`` :  params may be rewritten depending on dataset.
            - ``False``:  params may be changed only manually
                or with tuning.
        timer: Timer for Algo.

    """

    _default_params: Dict = {}
    optimization_search_space: Dict = {}
    # TODO: add checks here
    _fit_checks: Tuple = ()
    _transform_checks: Tuple = ()
    _params: Dict = None
    _name = "AbstractAlgo"

    @property
    def name(self) -> str:
        """Get model name."""
        return self._name

    @property
    def features(self) -> List[str]:
        """Get list of features."""
        return self._features

    @features.setter
    def features(self, val: Sequence[str]):
        """List of features."""
        self._features = list(val)

    @property
    def is_fitted(self) -> bool:
        """Get flag is the model fitted or not."""
        return self.features is not None

    @property
    def params(self) -> dict:
        """Get model's params dict."""
        if self._params is None:
            self._params = copy(self.default_params)
        return self._params

    @params.setter
    def params(self, new_params: dict):
        assert isinstance(new_params, dict)
        self._params = {**self.params, **new_params}

    def init_params_on_input(self, train_valid_iterator: TrainValidIterator) -> dict:
        """Init params depending on input data.

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Dict with model hyperparameters.

        """
        return self.params

    # TODO: Think about typing
    def __init__(
        self,
        default_params: Optional[dict] = None,
        freeze_defaults: bool = True,
        timer: Optional[TaskTimer] = None,
        optimization_search_space: Optional[dict] = {},
    ):
        self.task = None
        self.optimization_search_space = optimization_search_space

        self.freeze_defaults = freeze_defaults
        if default_params is None:
            default_params = {}

        self.default_params = {**self._default_params, **default_params}

        self.models = []
        self._features = None

        self.timer = timer
        if timer is None:
            self.timer = PipelineTimer().start().get_task_timer()

        self._nan_rate = None

    @abstractmethod
    def fit_predict(self, train_valid_iterator: TrainValidIterator) -> LAMLDataset:
        """Abstract method.

        Fit new algo on iterated datasets and predict on valid parts.

        Args:
            train_valid_iterator: Classic cv-iterator.

        """
        # self._features = train_valid_iterator.features

    @abstractmethod
    def predict(self, test: LAMLDataset) -> LAMLDataset:
        """Predict target for input data.

        Args:
            test: Dataset on test.

        Returns:
            Dataset with predicted values.

        """

    def score(self, dataset: LAMLDataset) -> float:
        """Score prediction on dataset with defined metric.

        Args:
            dataset: Dataset with ground truth and predictions.

        Returns:
            Metric value.

        """
        assert self.task is not None, "No metric defined. Should be fitted on dataset first."
        metric = self.task.get_dataset_metric()

        return metric(dataset, dropna=True)

    def set_prefix(self, prefix: str):
        """Set prefix to separate models from different levels/pipelines.

        Args:
            prefix: String with prefix.

        """
        self._name = "_".join([prefix, self._name])

    def set_timer(self, timer: TaskTimer) -> "MLAlgo":
        """Set timer."""
        self.timer = timer

        return self


class TabularMLAlgo(MLAlgo):
    """Machine learning algorithms that accepts numpy arrays as input."""

    _name: str = "TabularAlgo"

    def _set_prediction(self, dataset: NumpyDataset, preds_arr: np.ndarray) -> NumpyDataset:
        """Insert predictions to dataset with. Inplace transformation.

        Args:
            dataset: Dataset to transform.
            preds_arr: Array with predicted values.

        Returns:
            Transformed dataset.

        """
        prefix = "{0}_prediction".format(self._name)
        prob = self.task.name in ["binary", "multiclass"]
        dataset.set_data(preds_arr, prefix, NumericRole(np.float32, force_input=True, prob=prob))

        return dataset

    def fit_predict_single_fold(self, train: TabularDataset, valid: TabularDataset) -> Tuple[Any, np.ndarray]:
        """Train on train dataset and predict on holdout dataset.

        Args:
            train: Train Dataset.
            valid: Validation Dataset.

        Returns:  # noqa: DAR202
            Target predictions for valid dataset.

        """
        raise NotImplementedError

    def fit_predict(self, train_valid_iterator: TrainValidIterator) -> NumpyDataset:
        """Fit and then predict accordig the strategy that uses train_valid_iterator.

        If item uses more then one time it will
        predict mean value of predictions.
        If the element is not used in training then
        the prediction will be ``numpy.nan`` for this item

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Dataset with predicted values.

        """
        self.timer.start()

        assert self.is_fitted is False, "Algo is already fitted"
        # init params on input if no params was set before
        if self._params is None:
            self.params = self.init_params_on_input(train_valid_iterator)

        iterator_len = len(train_valid_iterator)
        if iterator_len > 1:
            logger.info("Start fitting \x1b[1m{}\x1b[0m ...".format(self._name))
            stop_params = ["cat_features", "cont_features", "cat_dims", "cat_vc"]
            printable_params = {key: value for key, value in self.params.items() if key not in stop_params}
            logger.debug(f"Training params: {printable_params}")

        # save features names
        self._features = train_valid_iterator.features
        # get metric and loss if None
        self.task = train_valid_iterator.train.task

        # get empty validation data to write prediction
        # TODO: Think about this cast
        preds_ds = cast(NumpyDataset, train_valid_iterator.get_validation_data().empty().to_numpy())

        outp_dim = 1
        if self.task.name == "multiclass":
            outp_dim = int(np.max(preds_ds.target) + 1)
        elif (self.task.name == "multi:reg") or (self.task.name == "multilabel"):
            outp_dim = preds_ds.target.shape[1]
        # save n_classes to infer params
        self.n_classes = outp_dim

        preds_arr = np.zeros((preds_ds.shape[0], outp_dim), dtype=np.float32)
        counter_arr = np.zeros((preds_ds.shape[0], 1), dtype=np.float32)

        # TODO: Make parallel version later
        for n, (idx, train, valid) in enumerate(train_valid_iterator):
            if iterator_len > 1:
                logger.info2(
                    "===== Start working with \x1b[1mfold {}\x1b[0m for \x1b[1m{}\x1b[0m =====".format(n, self._name)
                )
            self.timer.set_control_point()

            model, pred = self.fit_predict_single_fold(train, valid)
            self.models.append(model)
            preds_arr[idx] += pred.reshape((pred.shape[0], -1))
            counter_arr[idx] += 1

            self.timer.write_run_info()

            if (n + 1) != len(train_valid_iterator):
                # split into separate cases because timeout checking affects parent pipeline timer
                if self.timer.time_limit_exceeded():
                    logger.info("Time limit exceeded after calculating fold {0}\n".format(n))
                    break

        preds_arr /= np.where(counter_arr == 0, 1, counter_arr)
        preds_arr = np.where(counter_arr == 0, np.nan, preds_arr)

        preds_ds = self._set_prediction(preds_ds, preds_arr)

        if iterator_len > 1:
            logger.info(f"Fitting \x1b[1m{self._name}\x1b[0m finished. score = \x1b[1m{self.score(preds_ds)}\x1b[0m")

        if iterator_len > 1 or "Tuned" not in self._name:
            logger.info("\x1b[1m{}\x1b[0m fitting and predicting completed".format(self._name))
        return preds_ds

    def predict_single_fold(self, model: Any, dataset: TabularDataset) -> np.ndarray:
        """Implements prediction on single fold.

        Args:
            model: Model uses to predict.
            dataset: Dataset used for prediction.

        Returns:  # noqa: DAR202
            Predictions for input dataset.

        """
        raise NotImplementedError

    def predict(self, dataset: TabularDataset) -> NumpyDataset:
        """Mean prediction for all fitted models.

        Args:
            dataset: Dataset used for prediction.

        Returns:
            Dataset with predicted values.

        """
        assert self.models != [], "Should be fitted first."
        preds_ds = dataset.empty().to_numpy()
        preds_arr = None

        for model in self.models:
            if preds_arr is None:
                preds_arr = self.predict_single_fold(model, dataset)
            else:
                preds_arr += self.predict_single_fold(model, dataset)

        preds_arr /= len(self.models)
        preds_arr = preds_arr.reshape((preds_arr.shape[0], -1))
        preds_ds = self._set_prediction(preds_ds, preds_arr)

        return preds_ds



"""Image feaures."""

from typing import Any

import torch

from ...dataset.base import LAMLDataset
from ...image.utils import pil_loader
from ...transformers.base import ColumnsSelector
from ...transformers.base import LAMLTransformer
from ...transformers.base import SequentialTransformer
from ...transformers.base import UnionTransformer
from ...transformers.image import AutoCVWrap
from ...transformers.image import ImageFeaturesTransformer
from ...transformers.numeric import FillInf
from ...transformers.numeric import FillnaMedian
from ...transformers.numeric import StandardScaler
from ..utils import get_columns_by_role
from .base import FeaturesPipeline


class ImageDataFeatures:
    """Class contains basic features transformations for image data.

    Args:
        **kwargs: Default parameters.

    """

    def __init__(self, **kwargs: Any):
        self.hist_size = 30
        self.is_hsv = True
        self.n_jobs = 4
        self.loader = pil_loader

        self.embed_model = "efficientnet_b0.ra_in1k"
        self.weights_path = None
        self.subs = 10000
        self.cache_dir = "../cache_CV"
        self.device = torch.device("cuda:0") if torch.cuda.is_available() else torch.device("cpu")
        self.batch_size = 128
        self.verbose = True

        self.random_state = 42

        for k in kwargs:
            self.__dict__[k] = kwargs[k]


class ImageSimpleFeatures(FeaturesPipeline, ImageDataFeatures):
    """Class contains simple color histogram features for image data."""

    def create_pipeline(self, train: LAMLDataset) -> LAMLTransformer:
        """Create pipeline for images data.

        Args:
            train: Dataset with train features.

        Returns:
            Transformer.

        """
        transformers_list = []

        # process texts
        imgs = get_columns_by_role(train, "Path")
        if len(imgs) > 0:
            imgs_processing = SequentialTransformer(
                [
                    ColumnsSelector(keys=imgs),
                    ImageFeaturesTransformer(self.hist_size, self.is_hsv, self.n_jobs, self.loader),
                    SequentialTransformer([FillInf(), FillnaMedian(), StandardScaler()]),
                ]
            )
            transformers_list.append(imgs_processing)

        union_all = UnionTransformer(transformers_list)

        return union_all


class ImageAutoFeatures(FeaturesPipeline, ImageDataFeatures):
    """Class contains efficient-net embeddings features for image data."""

    def create_pipeline(self, train: LAMLDataset) -> LAMLTransformer:
        """Create pipeline for images data.

        Args:
            train: Dataset with train features.

        Returns:
            Transformer.

        """
        transformers_list = []
        # process texts
        imgs = get_columns_by_role(train, "Path")
        if len(imgs) > 0:
            imgs_processing = SequentialTransformer(
                [
                    ColumnsSelector(keys=imgs),
                    AutoCVWrap(
                        self.embed_model,
                        self.weights_path,
                        self.cache_dir,
                        self.subs,
                        self.device,
                        self.n_jobs,
                        self.random_state,
                        self.batch_size,
                        self.verbose,
                    ),
                    SequentialTransformer([FillInf(), FillnaMedian(), StandardScaler()]),
                ]
            )
            transformers_list.append(imgs_processing)

        union_all = UnionTransformer(transformers_list)

        return union_all



"""Metrics and loss functions for Torch based models."""

from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

import torch

from torch import nn

from .base import Loss


class TorchLossWrapper(nn.Module):
    """Customize PyTorch-based loss.

    Args:
        func: loss to customize. Example: `torch.nn.MSELoss`.
        **kwargs: additional parameters.

    Returns:
        callable loss, uses format (y_true, y_pred, sample_weight).

    """

    def __init__(self, func: Callable, flatten=False, log=False, **kwargs: Any):
        super(TorchLossWrapper, self).__init__()
        self.base_loss = func(reduction="none", **kwargs)
        self.flatten = flatten
        self.log = log

    def forward(
        self,
        y_true: torch.Tensor,
        y_pred: torch.Tensor,
        sample_weight: Optional[torch.Tensor] = None,
    ):
        """Forward-pass."""
        if self.flatten:
            y_true = y_true[:, 0].type(torch.int64)

        if self.log:
            y_pred = torch.log(y_pred)

        outp = self.base_loss(y_pred, y_true)

        if len(outp.shape) == 2:
            outp = outp.sum(dim=1)

        if sample_weight is not None:
            outp = outp * sample_weight
            return outp.mean() / sample_weight.mean()

        return outp.mean()


def torch_rmsle(
    y_true: torch.Tensor,
    y_pred: torch.Tensor,
    sample_weight: Optional[torch.Tensor] = None,
):
    """Computes Root Mean Squared Logarithmic Error.

    Args:
        y_true: true target values.
        y_pred: predicted target values.
        sample_weight: specify weighted mean.

    Returns:
        metric value.

    """
    y_pred = torch.log1p(y_pred)
    y_true = torch.log1p(y_true)

    outp = (y_pred - y_true) ** 2
    if len(outp.shape) == 2:
        outp = outp.sum(dim=1)

    if sample_weight is not None:
        outp = outp * sample_weight
        return outp.mean() / sample_weight.mean()

    return outp.mean()


def torch_quantile(
    y_true: torch.Tensor,
    y_pred: torch.Tensor,
    sample_weight: Optional[torch.Tensor] = None,
    q: float = 0.9,
):
    """Computes Mean Quantile Error.

    Args:
        y_true: true target values.
        y_pred: predicted target values.
        sample_weight: specify weighted mean.
        q: metric coefficient.

    Returns:
        metric value.

    """
    err = y_pred - y_true
    s = err < 0
    err = torch.abs(err)
    err = torch.where(s, err * (1 - q), err * q)

    if len(err.shape) == 2:
        err = err.sum(dim=1)

    if sample_weight is not None:
        err = err * sample_weight
        return err.mean() / sample_weight.mean()

    return err.mean()


def torch_fair(
    y_true: torch.Tensor,
    y_pred: torch.Tensor,
    sample_weight: Optional[torch.Tensor] = None,
    c: float = 0.9,
):
    """Computes Mean Fair Error.

    Args:
        y_true: true target values.
        y_pred: predicted target values.
        sample_weight: specify weighted mean.
        c: metric coefficient.

    Returns:
        metric value.

    """
    x = torch.abs(y_pred - y_true) / c
    err = c ** 2 * (x - torch.log(x + 1))

    if len(err.shape) == 2:
        err = err.sum(dim=1)

    if sample_weight is not None:
        err = err * sample_weight
        return err.mean() / sample_weight.mean()

    return err.mean()


def torch_huber(
    y_true: torch.Tensor,
    y_pred: torch.Tensor,
    sample_weight: Optional[torch.Tensor] = None,
    a: float = 0.9,
):
    """Computes Mean Huber Error.

    Args:
        y_true: true target values.
        y_pred: predicted target values.
        sample_weight: specify weighted mean.
        a: metric coefficient.

    Returns:
        metric value.

    """
    err = y_pred - y_true
    s = torch.abs(err) < a
    err = torch.where(s, 0.5 * (err ** 2), a * torch.abs(err) - 0.5 * (a ** 2))

    if len(err.shape) == 2:
        err = err.sum(dim=1)

    if sample_weight is not None:
        err = err * sample_weight
        return err.mean() / sample_weight.mean()

    return err.mean()


def torch_f1(
    y_true: torch.Tensor,
    y_pred: torch.Tensor,
    sample_weight: Optional[torch.Tensor] = None,
):
    """Computes F1 macro.

    Args:
        y_true: true target values.
        y_pred: predicted target values.
        sample_weight: specify weighted mean.

    Returns:
        metric value.

    """
    y_pred = torch.sigmoid(y_pred)
    y_true = y_true[:, 0].type(torch.int64)
    y_true_ohe = torch.zeros_like(y_pred)

    y_true_ohe[range(y_true.shape[0]), y_true] = 1
    tp = y_true_ohe * y_pred
    if sample_weight is not None:
        sample_weight = sample_weight.unsqueeze(-1)
        sm = sample_weight.mean()
        tp = (tp * sample_weight).mean(dim=0) / sm
        f1 = (2 * tp) / (
            (y_pred * sample_weight).mean(dim=0) / sm + (y_true_ohe * sample_weight).mean(dim=0) / sm + 1e-7
        )

        return -f1.mean()

    tp = torch.mean(tp, dim=0)

    f1 = (2 * tp) / (y_pred.mean(dim=0) + y_true_ohe.mean(dim=0) + 1e-7)

    f1[f1 != f1] = 0

    return -f1.mean()


def torch_mape(
    y_true: torch.Tensor,
    y_pred: torch.Tensor,
    sample_weight: Optional[torch.Tensor] = None,
):
    """Computes Mean Absolute Percentage Error.

    Args:
        y_true: true target values.
        y_pred: predicted target values.
        sample_weight: specify weighted mean.

    Returns:
        metric value.

    """
    err = (y_true - y_pred) / y_true
    err = torch.abs(err)

    if len(err.shape) == 2:
        err = err.sum(dim=1)

    if sample_weight is not None:
        err = err * sample_weight
        return err.mean() / sample_weight.mean()

    return err.mean()


_torch_loss_dict = {
    "mse": (nn.MSELoss, False, False),
    "mae": (nn.L1Loss, False, False),
    "logloss": (nn.BCEWithLogitsLoss, False, False),
    "crossentropy": (nn.CrossEntropyLoss, True, False),
    "rmsle": (torch_rmsle, False, False),
    "mape": (torch_mape, False, False),
    "quantile": (torch_quantile, False, False),
    "fair": (torch_fair, False, False),
    "huber": (torch_huber, False, False),
    "f1": (torch_f1, False, False),
}


class TORCHLoss(Loss):
    """Loss used for PyTorch.

    Args:
        loss: name or callable objective function.
        loss_params: additional loss parameters.

    """

    def __init__(self, loss: Union[str, Callable], loss_params: Optional[Dict] = None):
        self.loss_params = {}
        if loss_params is not None:
            self.loss_params = loss_params

        if loss in ["mse", "mae", "logloss", "crossentropy"]:
            self.loss = TorchLossWrapper(*_torch_loss_dict[loss], **self.loss_params)
        elif type(loss) is str:
            self.loss = partial(_torch_loss_dict[loss][0], **self.loss_params)
        else:
            self.loss = partial(loss, **self.loss_params)



"""Provides an internal interface for working with text features."""


__all__ = [
    "tokenizer",
    "dl_transformers",
    "sentence_pooling",
    "weighted_average_transformer",
    "embed_dataset",
]



"""Custom metrics and loss functions for Catboost."""

from typing import Callable

import numpy as np


# TODO: Calc metrics on gpu slow down warning. Check it


class CBCustomMetric:
    """Metric wrapper class for CatBoost.

    Args:
        metric: Callable metric.
        greater_is_better: Bool with metric direction.

    """

    def __init__(self, metric: Callable, greater_is_better: bool = True, bw_func: Callable = None):
        self.metric = metric
        self.greater_is_better = greater_is_better
        self._bw_func = bw_func

    @staticmethod
    def get_final_error(error, weight):  # noqa D102
        return error

    def is_max_optimal(self):  # noqa D102
        return self.greater_is_better

    def evaluate(self, approxes, target, weight):
        """Calculate metric."""
        raise NotImplementedError

    def _transform(self, approxes, target):
        target = np.array(target)
        pred = np.array(approxes[0])
        if self._bw_func is not None:
            target = self._bw_func(target)
            pred = self._bw_func(pred)

        return pred, target

    def _evaluate(self, approxes, target, weight):
        pred, target = self._transform(approxes, target)
        if weight is None:
            score = self.metric(target, pred)
            return score, 1
        else:
            try:
                score = self.metric(target, pred, sample_weight=np.array(weight))
            except TypeError:
                score = self.metric(target, pred)
            return score, weight


class CBRegressionMetric(CBCustomMetric):
    """Regression metric wrapper for CatBoost."""

    def evaluate(self, approxes, target, weight):
        """Calculate metric."""
        assert len(approxes) == 1
        assert len(approxes[0]) == len(target)
        assert weight is None or len(target) == len(weight)

        return self._evaluate(approxes, target, weight)


class CBClassificationMetric(CBCustomMetric):
    """Classification metric wrapper for CatBoost."""

    def __init__(
        self,
        metric: Callable,
        greater_is_better: bool,
        bw_func: Callable = None,
        use_proba: bool = True,
    ):
        super(CBClassificationMetric, self).__init__(metric, greater_is_better, bw_func)
        self.use_proba = use_proba

    def _transform(self, approxes, target):
        target = np.array(target)
        pred = np.array(approxes[0])
        if self._bw_func is not None:
            pred = self._bw_func(pred)

        pred = np.exp(pred)
        pred = pred / (1 + pred)
        if self.use_proba:
            return pred, target
        else:
            return np.round(pred), target

    def evaluate(self, approxes, target, weight):
        """Calculate metric."""
        assert len(approxes) == 1
        assert len(approxes[0]) == len(target)
        assert weight is None or len(target) == len(weight)

        return self._evaluate(approxes, target, weight)


class CBMulticlassMetric(CBCustomMetric):
    """Multiclassification metric wrapper for CatBoost."""

    def __init__(
        self,
        metric: Callable,
        greater_is_better: bool,
        bw_func: Callable = None,
        use_proba: bool = True,
    ):
        super().__init__(metric, greater_is_better, bw_func)
        self.use_proba = use_proba

    def _transform(self, approxes, target):
        target = np.array(target)
        pred = np.array(approxes)
        if self._bw_func is not None:
            pred = self._bw_func(pred)

        pred = np.exp(approxes)
        pred = pred / (1 + pred)
        if self.use_proba:
            return pred, target
        else:
            return np.argmax(pred, axis=0), target

    def evaluate(self, approxes, target, weight):
        """Calculate metric."""
        assert len(approxes[0]) == len(target)
        assert weight is None or len(target) == len(weight)

        return self._evaluate(approxes, target, weight)



"""Neural Net modules for different data types."""

import logging

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union
import numpy as np
import torch
import torch.nn as nn
from ..tasks.base import Task


logger = logging.getLogger(__name__)


class UniversalDataset:
    """Dataset class for mixed data.

    Args:
        data: Dict with data.
        y: Array of target variable.
        w: Optional array of observation weight.
        tokenizer: Transformers tokenizer.
        max_length: Max sentence length.
        stage: Name of current training / inference stage.

    """

    def __init__(
        self,
        data: Dict[str, np.ndarray],
        y: np.ndarray,
        w: Optional[np.ndarray] = None,
        tokenizer: Optional[Any] = None,
        max_length: int = 256,
        stage: str = "test",
    ):
        self.data = data
        self.y = y
        self.w = w
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.stage = stage

    def __len__(self) -> int:
        return len(self.y)

    def __getitem__(self, index: int) -> Dict[str, np.ndarray]:
        res = {"label": self.y[index]}
        res.update({key: value[index] for key, value in self.data.items() if key != "text"})
        if (self.tokenizer is not None) and ("text" in self.data):
            sent = self.data["text"][index, 0]  # only one column
            _split = sent.split("[SEP]")
            sent = _split if len(_split) == 2 else (sent,)
            data = self.tokenizer.encode_plus(
                *sent, add_special_tokens=True, max_length=self.max_length, padding="max_length", truncation=True
            )

            res.update({i: np.array(data[i]) for i in data.keys()})
        if self.w is not None:
            res["weight"] = self.w[index]

        return res


class Clump(nn.Module):
    """Clipping input tensor.

    Args:
        min_v: Min value.
        max_v: Max value.

    """

    def __init__(self, min_v: int = -50, max_v: int = 50):
        super(Clump, self).__init__()

        self.min_v = min_v
        self.max_v = max_v

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward-pass."""
        x = torch.clamp(x, self.min_v, self.max_v)
        return x


class TorchUniversalModel(nn.Module):
    """Mixed data model.

    Class for preparing input for DL model with mixed data.

    Args:
        loss: Callable torch loss with order of arguments (y_true, y_pred).
        task: Task object.
        torch_model: Torch model.
        n_out: Number of output dimensions.
        cont_embedder: Torch module for numeric data.
        cont_params: Dict with numeric model params.
        cat_embedder: Torch module for category data.
        cat_params: Dict with category model params.
        text_embedder: Torch module for text data.
        text_params: Dict with text model params.
        loss_on_logits: Calculate loss on logits or on predictions of model for classification tasks.
        bias: Array with last hidden linear layer bias.

    """

    def __init__(
        self,
        loss: Callable,
        task: Task,
        torch_model: nn.Module,
        n_out: int = 1,
        cont_embedder_: Optional[Any] = None,
        cont_params: Optional[Dict] = None,
        cat_embedder_: Optional[Any] = None,
        cat_params: Optional[Dict] = None,
        text_embedder: Optional[Any] = None,
        text_params: Optional[Dict] = None,
        loss_on_logits: bool = True,
        bias: Union[np.array, torch.Tensor] = None,
        **kwargs,
    ):
        super(TorchUniversalModel, self).__init__()
        self.n_out = n_out
        self.loss = loss
        self.task = task
        self.loss_on_logits = loss_on_logits

        self.cont_embedder = None
        self.cat_embedder = None
        self.text_embedder = None

        n_in = 0
        if cont_embedder_ is not None:
            self.cont_embedder = cont_embedder_(**cont_params)
            n_in += self.cont_embedder.get_out_shape()
        if cat_embedder_ is not None:
            self.cat_embedder = cat_embedder_(**cat_params)
            n_in += self.cat_embedder.get_out_shape()
        if text_embedder is not None:
            self.text_embedder = text_embedder(**text_params)
            n_in += self.text_embedder.get_out_shape()

        self.torch_model = (
            torch_model(
                **{
                    **kwargs,
                    **{
                        "n_in": n_in,
                        "n_out": n_out,
                        "loss": loss,
                        "task": task,
                    },
                }
            )
            if torch_model is not None
            else nn.Sequential(nn.Linear(n_in, n_out))
        )

        if bias is not None:
            self._set_last_layer(self.torch_model, bias)

        self.сlump = Clump()
        self.sig = nn.Sigmoid()
        self.softmax = nn.Softmax(dim=1)

    def _set_last_layer(self, torch_model, bias):
        use_skip = getattr(torch_model, "use_skip", False)
        self._init_last_layers(torch_model, bias, use_skip)

    def _init_last_layers(self, torch_model, bias, use_skip=False):
        try:
            all_layers = list(torch_model.children())
            layers = list(
                filter(
                    lambda x: isinstance(x, nn.Linear) or isinstance(x, nn.Sequential),
                    all_layers,
                )
            )
            if len(layers) == 0:
                last_layer = all_layers[-1]
                self._set_last_layer(last_layer, bias)

            else:
                last_layer = layers[-1]
                while isinstance(last_layer, nn.Sequential):
                    last_layer = list(
                        filter(lambda x: isinstance(x, nn.Linear) or isinstance(x, nn.Sequential), last_layer)
                    )[-1]
                bias = torch.Tensor(bias)
                last_layer.bias.data = bias
                shape = last_layer.weight.data.shape
                last_layer.weight.data = torch.zeros(shape[0], shape[1], requires_grad=True)
            if use_skip:
                if len(layers) <= 1:
                    last_layer = all_layers[-2]
                    self._set_last_layer(last_layer, bias)
                else:
                    pre_last_layer = layers[-2]
                    while isinstance(last_layer, nn.Sequential):
                        pre_last_layer = list(
                            filter(lambda x: isinstance(x, nn.Linear) or isinstance(x, nn.Sequential), pre_last_layer)
                        )[-1]
                    bias = torch.Tensor(bias)
                    pre_last_layer.bias.data = bias
                    shape = pre_last_layer.weight.data.shape
                    pre_last_layer.weight.data = torch.zeros(shape[0], shape[1], requires_grad=True)
        except:
            logger.info3("Last linear layer not founded, so init_bias=False")

    def get_logits(self, inp: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Forward-pass of model with embeddings."""
        outputs = []
        if self.cont_embedder is not None:
            outputs.append(self.cont_embedder(inp))

        if self.cat_embedder is not None:
            outputs.append(self.cat_embedder(inp))

        if self.text_embedder is not None:
            outputs.append(self.text_embedder(inp))

        if len(outputs) > 1:
            output = torch.cat(outputs, dim=1)
        else:
            output = outputs[0]

        logits = self.torch_model(output)
        return logits

    def get_preds_from_logits(self, logits: torch.Tensor) -> torch.Tensor:
        """Prediction from logits."""
        if self.task.name in ["binary", "multilabel"]:
            out = self.sig(self.сlump(logits))
        elif self.task.name == "multiclass":
            # can't find self.clump when predicting
            out = self.softmax(torch.clamp(logits, -50, 50))
        else:
            out = logits

        return out

    def forward(self, inp: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Forward-pass with output loss."""
        x = self.get_logits(inp)
        if not self.loss_on_logits:
            x = self.get_preds_from_logits(x)

        loss = self.loss(inp["label"].view(inp["label"].shape[0], -1), x, inp.get("weight", None))
        return loss

    def predict(self, inp: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Prediction."""
        x = self.get_logits(inp)
        x = self.get_preds_from_logits(x)
        return x



"""Whitebox features."""

import numpy as np

from ...dataset.np_pd_dataset import PandasDataset
from ...dataset.roles import NumericRole
from ...transformers.base import ColumnsSelector
from ...transformers.base import LAMLTransformer
from ...transformers.base import UnionTransformer
from ..utils import get_columns_by_role
from .base import FeaturesPipeline
from .base import TabularDataFeatures


class WBFeatures(FeaturesPipeline, TabularDataFeatures):
    """Simple WhiteBox pipeline.

    Just handles dates, other are handled inside WhiteBox.

    """

    def create_pipeline(self, train: PandasDataset) -> LAMLTransformer:
        """Create pipeline for WhiteBox.

        Args:
            train: Dataset with train features.

        Returns:
            Transformer.

        """
        others = get_columns_by_role(train, "Category") + get_columns_by_role(train, "Numeric")

        transformer_list = [
            self.get_datetime_diffs(train),
            self.get_datetime_seasons(train, NumericRole(np.float32)),
            ColumnsSelector(others),
        ]

        # final pipeline
        union_all = UnionTransformer([x for x in transformer_list if x is not None])

        return union_all



"""Neural net for tabular datasets."""


from lightautoml.utils.installation import __validate_extra_deps


__validate_extra_deps("nlp")


import gc
import logging
import os
import sys
import uuid

from copy import copy
from typing import Dict

import numpy as np
import optuna
import pandas as pd
import torch
import torch.nn as nn

from torch.optim.lr_scheduler import ReduceLROnPlateau

from ..dataset.np_pd_dataset import NumpyDataset
from ..tasks.losses.torch import TorchLossWrapper
from ..utils.installation import __validate_extra_deps
from ..validation.base import TrainValidIterator


__validate_extra_deps("nlp")

try:
    from transformers import AutoTokenizer

    from ..pipelines.features.text_pipeline import _model_name_by_lang
except:
    import warnings

    warnings.warn("'transformers' - package isn't installed")

from ..ml_algo.base import TabularDataset
from ..ml_algo.base import TabularMLAlgo
from ..pipelines.utils import get_columns_by_role
from ..text.embed import (
    BasicCatEmbeddingFlat,
    CatEmbedder,
    DenseEmbedding,
    DenseEmbeddingFlat,
    LinearEmbedding,
    LinearEmbeddingFlat,
    PLREmbedding,
    PLREmbeddingFlat,
    SoftEmbedding,
    SoftEmbeddingFlat,
    WeightedCatEmbedding,
    BasicCatEmbedding,
    WeightedCatEmbeddingFlat,
)
from ..text.embed import ContEmbedder
from ..text.embed import TextBert
from ..text.nn_model import TorchUniversalModel
from ..text.nn_model import UniversalDataset
from ..text.trainer import Trainer
from ..text.utils import collate_dict
from ..text.utils import inv_sigmoid
from ..text.utils import inv_softmax
from ..text.utils import is_shuffle
from ..text.utils import parse_devices
from ..text.utils import seed_everything
from .torch_based.nn_models import MLP, TabNet
from .torch_based.nn_models import NODE
from .torch_based.nn_models import SNN
from .torch_based.nn_models import DenseLightModel
from .torch_based.nn_models import DenseModel
from .torch_based.nn_models import LinearLayer
from .torch_based.nn_models import ResNetModel
from .torch_based.nn_models import _LinearLayer
from .torch_based.nn_models import AutoInt
from .torch_based.nn_models import FTTransformer


logger = logging.getLogger(__name__)

model_by_name = {
    "denselight": DenseLightModel,
    "dense": DenseModel,
    "resnet": ResNetModel,
    "mlp": MLP,
    "linear_layer": LinearLayer,
    "_linear_layer": _LinearLayer,
    "snn": SNN,
    "node": NODE,
    "autoint": AutoInt,
    "tabnet": TabNet,
    "fttransformer": FTTransformer,
}
input_type_by_name = {
    "denselight": "flat",
    "dense": "flat",
    "resnet": "flat",
    "mlp": "flat",
    "linear_layer": "flat",
    "_linear_layer": "flat",
    "snn": "flat",
    "node": "flat",
    "autoint": "seq",
    "tabnet": "flat",
    "fttransformer": "seq",
}
cat_embedder_by_name_flat = {
    "cat": CatEmbedder,
    "cat_no_dropout": BasicCatEmbeddingFlat,
    "weighted": WeightedCatEmbeddingFlat,
}
cat_embedder_by_name = {
    "cat_no_dropout": BasicCatEmbedding,
    "cat_no_dropout": BasicCatEmbedding,
    "weighted": WeightedCatEmbedding,
}

cont_embedder_by_name_flat = {
    "cont": ContEmbedder,
    "linear": LinearEmbeddingFlat,
    "dense": DenseEmbeddingFlat,
    "plr": PLREmbeddingFlat,
    "soft": SoftEmbeddingFlat,
}
cont_embedder_by_name = {
    "cont": LinearEmbedding,
    "linear": LinearEmbedding,
    "dense": DenseEmbedding,
    "plr": PLREmbedding,
    "soft": SoftEmbedding,
}


class TorchModel(TabularMLAlgo):
    """Neural net for tabular datasets.

    default_params:

        - bs: Batch size.
        - num_workers: Number of threads for multiprocessing.
        - max_length: Max sequence length.
        - opt_params: Dict with optim params.
        - scheduler_params: Dict with scheduler params.
        - is_snap: Use snapshots.
        - snap_params: Dict with SE parameters.
        - init_bias: Init last linear bias by mean target values.
        - n_epochs: Number of training epochs.
        - input_bn: Use 1d batch norm for input data.
        - emb_dropout: Dropout probability.
        - emb_ratio: Ratio for embedding size = (x + 1) // emb_ratio.
        - max_emb_size: Max embedding size.
        - device: Torch device or str.
        - use_cont: Use numeric data.
        - use_cat: Use category data.
        - use_text: Use text data.
        - lang: Text language.
        - bert_name: Name of HuggingFace transformer model.
        - pooling: Type of pooling strategy for bert model.
        - deterministic: CUDNN backend.
        - multigpu: Use Data Parallel.
        - model_with_emb: Use model with custom embeddings.
        - loss: Torch loss or str or func with (y_pred, y_true) args.
        - loss_params: Dict with loss params.
        - loss_on_logits: Calculate loss on logits or on predictions of model for classification tasks.
        - clip_grad: Clip gradient before loss backprop.
        - clip_grad_params: Dict with clip_grad params.
        - dataset: Class for data retrieval
        - tuned: Tune custom model
        - num_init_features: Scale input dimension to another one
        - use_noise: Use Noise
        - use_bn: Use BatchNorm
        - path_to_save: Path to save model checkpoints,
          ``None`` - stay in memory.
        - random_state: Random state to take subsample.
        - verbose_inside: Number of steps between
          verbose inside epoch or ``None``.
        - verbose: Verbose every N epochs.

    freeze_defaults:

        - ``True`` :  params can be rewritten depending on dataset.
        - ``False``:  params can be changed only manually or with tuning.

    timer: :class:`~lightautoml.utils.timer.Timer` instance or ``None``.

    """

    _name: str = "TorchNN"
    _params: Dict = None

    _default_models_params = {
        "n_out": None,
        "hid_factor": [2, 2],
        "hidden_size": [512, 512, 512],
        "block_config": [2, 2],
        "compression": 0.5,
        "growth_size": 256,
        "bn_factor": 2,
        "drop_rate": 0.1,
        "noise_std": 0.05,
        "num_init_features": None,
        "act_fun": nn.ReLU,
        "use_noise": False,
        "use_bn": True,
    }

    _default_params = {
        "num_workers": 0,
        "pin_memory": False,
        "max_length": 256,
        "is_snap": False,
        "input_bn": False,
        "max_emb_size": 256,
        "bert_name": None,
        "pooling": "cls",
        "device": torch.device("cuda:0"),
        "use_cont": True,
        "use_cat": True,
        "use_text": True,
        "lang": "en",
        "deterministic": True,
        "multigpu": False,
        "random_state": 42,
        "model": "dense",
        "model_with_emb": False,
        "path_to_save": os.path.join("./models/", "model"),
        "verbose_inside": None,
        "verbose": 1,
        "n_epochs": 20,
        "snap_params": {"k": 3, "early_stopping": True, "patience": 16, "swa": True},
        "bs": 512,
        "emb_dropout": 0.1,
        "emb_ratio": 3,
        "opt": torch.optim.Adam,
        "opt_params": {"weight_decay": 0, "lr": 3e-4},
        "sch": ReduceLROnPlateau,
        "scheduler_params": {"patience": 10, "factor": 1e-2, "min_lr": 1e-5},
        "loss": None,
        "loss_params": {},
        "loss_on_logits": True,
        "clip_grad": False,
        "clip_grad_params": {},
        "init_bias": True,
        "dataset": UniversalDataset,
        "tuned": False,
        "optimization_search_space": None,
        "verbose_bar": False,
        "freeze_defaults": False,
        **_default_models_params,
    }

    def _infer_params(self):
        if self.params["path_to_save"] is not None:
            self.path_to_save = os.path.relpath(self.params["path_to_save"])
            if not os.path.exists(self.path_to_save):
                os.makedirs(self.path_to_save)
        else:
            self.path_to_save = None

        params = copy(self.params)
        loss = params["loss"]

        if isinstance(loss, str):
            loss = getattr(nn, loss)
        if loss is not None and issubclass(loss, nn.Module):
            loss = TorchLossWrapper(loss, **params["loss_params"])
        if loss is None:
            loss = self.task.losses["torch"].loss

        params["loss"] = loss
        params["metric"] = self.task.losses["torch"].metric_func

        if params["bert_name"] is None and params["use_text"]:
            params["bert_name"] = _model_name_by_lang[params.get("lang", "en")]

        is_text = (len(params["text_features"]) > 0) and (params["use_text"]) and (params["device"].type == "cuda")
        is_cat = (len(params["cat_features"]) > 0) and (params["use_cat"])
        is_cont = (len(params["cont_features"]) > 0) and (params["use_cont"])

        torch_model = params["model"]

        if isinstance(torch_model, str):
            assert torch_model in model_by_name, "Wrong model name. Available models: " + str(model_by_name.keys())
            if torch_model == "snn":
                params["init_bias"] = False
            torch_model = model_by_name[torch_model]

        assert issubclass(torch_model, nn.Module), "Wrong model format, only support torch models"

        # str to nn modules
        for p_name, module in [
            ["act_fun", nn],
            ["dataset", sys.modules[__name__]],
            ["opt", torch.optim],
            ["sch", torch.optim.lr_scheduler],
        ]:
            if isinstance(params[p_name], str):
                params[p_name] = getattr(module, params[p_name])

        # params = self._select_params(params)
        model = Trainer(
            net=TorchUniversalModel if not params["model_with_emb"] else params["model"],
            net_params={
                "task": self.task,
                "cont_embedder_": cont_embedder_by_name.get(params["cont_embedder"], LinearEmbedding)
                if input_type_by_name.get(params["model"], "flat") == "seq" and is_cont
                else cont_embedder_by_name_flat.get(params["cont_embedder"], ContEmbedder)
                if is_cont
                else None,
                "cont_params": {
                    # "num_dims": params["num_dims"],
                    # "input_bn": params["input_bn"],
                    # "device": params["device"],
                    # "embedding_size": params["embedding_size"],
                    **params
                }
                if is_cont
                else None,
                "cat_embedder_": cat_embedder_by_name.get(params["cat_embedder"], BasicCatEmbedding)
                if input_type_by_name.get(params["model"], "flat") == "seq" and is_cat
                else cat_embedder_by_name_flat.get(params["cat_embedder"], CatEmbedder)
                if is_cat
                else None,
                "cat_params": {
                    # "cat_vc": params["cat_vc"],
                    # "cat_dims": params["cat_dims"],
                    # "emb_dropout": params["emb_dropout"],
                    # "emb_ratio": params["emb_ratio"],
                    # "max_emb_size": params["max_emb_size"],
                    # "embedding_size": params["embedding_size"],
                    # "device": params["device"],
                    **params
                }
                if is_cat
                else None,
                "text_embedder": TextBert if is_text else None,
                "text_params": {
                    "model_name": params["bert_name"],
                    "pooling": params["pooling"],
                }
                if is_text
                else None,
                "torch_model": torch_model,
                **params,
            },
            **{"apex": False, **params},
        )

        self.train_params = {
            "dataset": params["dataset"],
            "bs": params["bs"],
            "num_workers": params["num_workers"],
            "pin_memory": params["pin_memory"],
            "tokenizer": AutoTokenizer.from_pretrained(params["bert_name"], use_fast=False) if is_text else None,
            "max_length": params["max_length"],
        }

        return model

    @staticmethod
    def get_mean_target(target, task_name: str):
        """Get target mean / inverse sigmoid transformation \
            to init bias in last layer of network.

        Args:
            target: Target values.
            task_name: One of the available task names

        Returns:
            Array with bias values.

        """
        if isinstance(target, pd.Series) or isinstance(target, pd.DataFrame):
            target = target.values
        target = target.reshape(target.shape[0], -1)
        bias = (
            np.nanmean(target, axis=0).astype(float)
            if (task_name != "multiclass")
            else np.unique(target, return_counts=True)[1]
        )
        bias = (
            inv_sigmoid(bias)
            if (task_name == "binary") or (task_name == "multilabel")
            else inv_softmax(bias)
            if (task_name == "multiclass")
            else bias
        )

        bias[bias == np.inf] = np.nanmax(bias[bias != np.inf])
        bias[bias == -np.inf] = np.nanmin(bias[bias != -np.inf])
        bias[bias == np.NaN] = np.nanmean(bias[bias != np.NaN])

        return bias

    def _init_params_on_input(self, train_valid_iterator) -> dict:
        """Init params that common for all folds.

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Dict with data parameters.

        """
        params = self.params
        new_params = {}
        new_params["device"], new_params["device_ids"] = parse_devices(params["device"], params["multigpu"])

        task_name = train_valid_iterator.train.task.name
        target = train_valid_iterator.train.target

        if params["n_out"] is None:
            new_params["n_out"] = 1 if task_name != "multiclass" else (np.max(target) + 1).astype(int)
            new_params["n_out"] = target.shape[1] if task_name in ["multi:reg", "multilabel"] else new_params["n_out"]

        cat_dims = []
        new_params["cat_features"] = get_columns_by_role(train_valid_iterator.train, "Category")

        # Cat_features are needed to be preprocessed with LE, where 0 = not known category
        valid = train_valid_iterator.get_validation_data()
        cat_value_counts = []
        for cat_feature in new_params["cat_features"]:
            num_unique_categories = (
                max(
                    max(train_valid_iterator.train[:, cat_feature].data),
                    max(valid[:, cat_feature].data),
                )
                + 1
            )
            if params["cat_embedder"] == "weighted":
                values, counts = np.unique(
                    np.concatenate([train_valid_iterator.train[:, cat_feature].data, valid[:, cat_feature].data]),
                    return_counts=True,
                )
                cat_value_counts.append(dict(zip(values, counts)))
            cat_dims.append(num_unique_categories)
        new_params["cat_dims"] = cat_dims
        new_params["cat_vc"] = cat_value_counts
        new_params["cont_features"] = get_columns_by_role(train_valid_iterator.train, "Numeric")
        new_params["num_dims"] = len(new_params["cont_features"])

        new_params["text_features"] = get_columns_by_role(train_valid_iterator.train, "Text")
        new_params["bias"] = self.get_mean_target(target, task_name) if params["init_bias"] else None

        logger.debug(f'number of text features: {len(new_params["text_features"])} ')
        logger.debug(f'number of categorical features: {len(new_params["cat_features"])} ')
        logger.debug(f'number of continuous features: {new_params["num_dims"]} ')

        return new_params

    def init_params_on_input(self, train_valid_iterator) -> dict:
        """Get model parameters depending on dataset parameters.

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Parameters of model.

        """
        suggested_params = copy(self.params)
        if self.freeze_defaults:
            # if user change defaults manually - keep it
            return suggested_params

        rows_num = len(train_valid_iterator.train) + len(train_valid_iterator.get_validation_data())
        bs = 256

        if rows_num > 50_000:
            bs = 512
        if rows_num > 100_000:
            bs = 1024
        if rows_num > 1_000_000:
            bs = 2048

        suggested_params["bs"] = bs
        suggested_params["snap_params"] = (
            {"k": 1, "early_stopping": True, "patience": 16, "swa": False}
            if self.params["model"] == "snn"
            else self.params["snap_params"]
        )
        return suggested_params

    def get_dataloaders_from_dicts(self, data_dict: Dict):
        """Construct dataloaders depending on stage.

        Args:
            data_dict: Dict with (stage_name, data) (key, value).

        Returns:
            Dataloaders.

        """
        datasets = {}
        for stage, value in data_dict.items():
            data = {
                name: value.data[cols].values
                for name, cols in zip(
                    ["text", "cat", "cont"],
                    [
                        self.params["text_features"],
                        self.params["cat_features"],
                        self.params["cont_features"],
                    ],
                )
                if len(cols) > 0
            }

            datasets[stage] = self.train_params["dataset"](
                data=data,
                y=value.target.values if stage != "test" else np.ones(len(value.data)),
                w=value.weights.values if value.weights is not None else np.ones(len(value.data)),
                tokenizer=self.train_params["tokenizer"],
                max_length=self.train_params["max_length"],
                stage=stage,
            )

        dataloaders = {
            stage: torch.utils.data.DataLoader(
                datasets[stage],
                batch_size=self.train_params["bs"],
                shuffle=is_shuffle(stage),
                num_workers=self.train_params["num_workers"],
                collate_fn=collate_dict,
                pin_memory=self.train_params["pin_memory"],
                drop_last={"train": True}.get(stage, False),
            )
            for stage, value in data_dict.items()
        }
        return dataloaders

    def fit_predict(self, train_valid_iterator: TrainValidIterator) -> NumpyDataset:
        """Fit and then predict accordig the strategy that uses train_valid_iterator.

        If item uses more then one time it will
        predict mean value of predictions.
        If the element is not used in training then
        the prediction will be ``numpy.nan`` for this item

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Dataset with predicted values.

        """
        if self._params is None:
            self.params = self.init_params_on_input(train_valid_iterator)
        self.params = self._init_params_on_input(train_valid_iterator)
        return super().fit_predict(train_valid_iterator)

    def fit_predict_single_fold(self, train, valid):
        """Implements training and prediction on single fold.

        Args:
            train: Train Dataset.
            valid: Validation Dataset.

        Returns:
            Tuple (model, predicted_values).

        """
        seed_everything(self.params["random_state"], self.params["deterministic"])
        task_name = train.task.name
        target = train.target
        self.params["bias"] = self.get_mean_target(target, task_name) if self.params["init_bias"] else None

        model = self._infer_params()

        model_path = (
            os.path.join(self.path_to_save, f"{uuid.uuid4()}.pickle") if self.path_to_save is not None else None
        )
        # init datasets
        dataloaders = self.get_dataloaders_from_dicts({"train": train.to_pandas(), "val": valid.to_pandas()})

        val_pred = model.fit(dataloaders)

        if model_path is None:
            model_path = model.state_dict(model_path)
        else:
            model.state_dict(model_path)

        model.clean()
        del dataloaders, model
        gc.collect()
        torch.cuda.empty_cache()
        return model_path, val_pred

    def predict_single_fold(self, model: any, dataset: TabularDataset) -> np.ndarray:
        """Predict target values for dataset.

        Args:
            model: Neural net object or dict or str.
            dataset: Test dataset.

        Returns:
            Predicted target values.

        """
        seed_everything(self.params["random_state"], self.params["deterministic"])
        dataloaders = self.get_dataloaders_from_dicts({"test": dataset.to_pandas()})

        if isinstance(model, (str, dict)):
            model = self._infer_params().load_state(model)

        pred = model.predict(dataloaders["test"], "test")

        model.clean()
        del dataloaders, model
        gc.collect()
        torch.cuda.empty_cache()

        return pred

    def _get_default_search_spaces(self, suggested_params: Dict, estimated_n_trials: int) -> Dict:
        def sample(estimated_n_trials: int, trial: optuna.trial.Trial, suggested_params: Dict) -> Dict:
            """Implements simple tuning sampling strategy.

            Args:
                trial: Current optuna Trial.
                estimated_n_trials: Estimated trials based on time spent on previous ones.
                suggested_params: Suggested params

            Returns:
                Dict with Sampled params.

            """
            # optionally
            trial_values = copy(suggested_params)

            trial_values["bs"] = trial.suggest_categorical("bs", [2 ** i for i in range(6, 11)])

            weight_decay_bin = trial.suggest_categorical("weight_decay_bin", [0, 1])
            if weight_decay_bin == 0:
                weight_decay = 0
            else:
                weight_decay = trial.suggest_float("weight_decay", low=1e-6, high=1e-2, log=True)

            lr = trial.suggest_float("lr", low=1e-5, high=1e-1, log=True)
            trial_values["opt_params"] = {
                "lr": lr,
                "weight_decay": weight_decay,
            }
            return trial_values

        return sample



"""Report generators and templates."""

from lightautoml.utils.installation import __validate_extra_deps

from .report_deco import ReportDeco
from .report_deco import ReportDecoNLP
from .report_deco import ReportDecoWhitebox


__validate_extra_deps("pdf")


__all__ = ["ReportDeco", "ReportDecoWhitebox", "ReportDecoNLP"]



import logging
import os

from html import escape
from numbers import Number
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import numpy as np
import pandas as pd


try:
    import gensim
except:
    import warnings

    warnings.warn("'gensim' - package isn't installed")

import torch
import torch.nn as nn

from torch.optim import Adam
from torch.optim.lr_scheduler import ReduceLROnPlateau
from tqdm import tqdm

from ...pipelines.features.text_pipeline import _tokenizer_by_lang
from ...text.utils import seed_everything
from .data_process import get_embedding_matrix
from .data_process import get_len_dataloader
from .data_process import get_len_dataset
from .data_process import get_tokenized
from .data_process import get_vocab
from .data_process import map_tokenized_to_id
from .l2x_model import L2XModel
from .utils import WrappedTokenizer
from .utils import WrappedVocabulary
from .utils import cross_entropy_multiple_class
from .utils import draw_html


logger = logging.getLogger(__name__)


class L2XTextExplainer:
    """
    Learning-to-Explain method for explaining NLP models for classification and regression,
    selecting the most informative (important) tokens in document.

    This method optimizes the variational lower bound on mutual information
    between target and selected number of tokens. Number of selected
    tokens is constant (n_important).

    The main differences from original model are:

    1. Dynamic batching, using the length of tokenized document.
    Without it the model can highlight unknown tokens (<UNK>).

    2. There are original and softsub methods for sampling important tokens.

    3. Additional regularizer-like term to encourage highliting neighborhood tokens.

    More precisely explanations of how the methods work in:
    1) Learning to Explain: An Information-Theoretic
    Perspective on Model Interpretation, J. Chen et al.
    https://arxiv.org/abs/1802.07814

    2) Reparameterizable Subset Sampling via Continuous Relaxations, S. Xie, S. Ermon.
    https://arxiv.org/abs/1901.10517

    Additional info:

    1. After traning all models will be on cpu and in evaluation mode.


    How should it works:

        >>> task = Task('reg')
        >>> automl = TabularNLPAutoML(task=task,
        >>>     timeout=600, gpu_ids = '0',
        >>>     general_params = {'nested_cv': False, 'use_algos': [['nn']]},
        >>>     text_params = {'lang': 'ru'})
        >>> automl.fit_predict(train, roles=roles)
        >>> from lightautoml.addons.interpretation import L2XTextExplainer
        >>> l2x = L2XTextExlpainer(automl, tokenizer='ru', device='cuda:0', n_important=10)
        >>> l2x.fit(train2, cols_to_explain='col1') # there can be also list of columns
        >>> explanations = l2x['col1'].explain_instances(data, batch_size=1)
        >>> explanations.get_all() # return list of (tokens_i, mask_i). The mask and tokens arrays have same sizes.
        >>> explanations[i].visualize_in_notebook() # produces the visualization of highlited tokens in document i.


    Args:
        automl: Automl object.
        tokenizer: String with language ['ru', 'en'],
            or callable function that take something,
            that deal with sentence in string to list of tokens.
            of list of strings. If None the lang
            from automl's text_params will be used.
        train_device: Device that will be used for traning L2X.
            Name of device should be valid for torch.device.
        inference_device: Device that will be used for inference L2X.
            Name of device should be valid for torch.device.
        verbose: Verbose training.
        binning_mode: For dynamic batching we use binning sampler by number
            of tokens in the tokenized document. This parameter specifies
            the method for constructing the boundaries for binning.
            Valid parameter values: 'linear' (min-max binning,
            like linspace), 'hist' (histogram binning).
        bins_number: Number of bins.
        n_important: Number of imembeddportant tokens.
        learning_rate: Learning rate of optimizer for traning L2X.
        n_epochs: Number of epochs for training L2X.
        optimizer: Should be optimizer in pytorch format.
        optimizer_params: Additional params of optimizer,
            exclude learning_rate.
        patience: Number of epochs before reducing learning rate.
        extreme_patience: Early stopping epochs.
        train_batch_size: Size of batch for training process.
        valid_batch_size: Size of batch for validation process.
        temperature: Temperature of concrete distribution sampling.
        temp_anneal_factor: Annealing temperature. The temperature will be
            multiplied by this coefficient after every epoch.
        conv_filters: Number of convolution kernels in model,
            that produces important tokens.
        conv_ksize: Size of convolution kernel.
        hidden_dim: Size of fully connected layer in L2X.
        drop_rate: Dropout rates in L2X.
        importance_sampler: Specifices method of sampling importance.
        embedder: Embedding dictionary or path to fasttext/dict of embeddings.
        embedding_dim: Dimention of embeddings.
        trainable_embeds: To train embeddings of L2X.
        max_vocab_length: Maximum vocabulary length. If -1 then include all in train set.
        gamma: Special coefficient, that encourage neighborhood of important tokens.
        gamma_anneal_factor: Annealing gamma. The gamma will be
            multiplied by this coefficient after every epoch.
        random_seed: Random seed.
        deterministic: Use cuda deterministic.
        cache_dir: Directory used for checkpointing model for early stopping.
            By default, it will infer from automl cache directory,
            or './l2x_cache' in case there is no opportunity to infer.

    """

    def __init__(
        self,
        automl,
        tokenizer: Optional[Union[str, Callable]] = None,
        train_device: str = "cpu",
        inference_device: str = "cpu",
        verbose: bool = True,
        binning_mode: str = "linear",
        bins_number: int = 3,
        n_important: int = 10,
        learning_rate: float = 1e-4,
        n_epochs: int = 200,
        optimizer: Type[torch.optim.Optimizer] = Adam,
        optimizer_params: Optional[Dict[str, Any]] = None,
        patience: int = 0,
        extreme_patience: int = 0,
        train_batch_size: int = 64,
        valid_batch_size: int = 16,
        temperature: float = 2,
        temp_anneal_factor: float = 0.985,
        conv_filters: int = 100,
        conv_ksize: int = 3,
        hidden_dim: int = 100,
        drop_rate: float = 0.2,
        importance_sampler: str = "softsub",
        embedder: Optional[Union[str, Dict[str, np.ndarray]]] = None,
        embedding_dim: Optional[int] = None,
        trainable_embeds: bool = False,
        max_vocab_length: int = -1,
        gamma: float = 0.01,
        gamma_anneal_factor: float = 1.0,
        random_seed: int = 42,
        deterministic: bool = True,
        cache_dir: Optional[str] = None,
    ):
        self.automl = automl
        self.reader = automl.reader
        self.task_name = automl.reader.task.name
        self.roles = automl.reader.roles

        if self.task_name == "binary":
            self._loss = nn.BCELoss()
            self.n_outs = 1
        elif self.task_name == "reg":
            self._loss = nn.MSELoss()
            self.n_outs = 1
        elif self.task_name == "multiclass":
            self._loss = cross_entropy_multiple_class
            self.n_outs = self.reader._n_classes

        if isinstance(tokenizer, str):
            if tokenizer not in ["ru", "en"]:
                raise ValueError("Tokenizer must be one 'ru' or 'en', but {} given".format(tokenizer))
            self._tokenizer = _tokenizer_by_lang[tokenizer](is_stemmer=False)
            self.tokenizer = WrappedTokenizer(self._tokenizer)
        elif tokenizer is None:
            # check is_stemmer=True or False,
            # who's the best and what easy to implement
            lang = automl.text_params["lang"]
            self._tokenizer = _tokenizer_by_lang[lang](is_stemmer=False)
            self.tokenizer = WrappedTokenizer(self._tokenizer)
        elif callable(tokenizer):
            self.tokenizer = tokenizer
        else:
            raise TypeError("Unkown type of tokenizer: {}".format(type(tokenizer)))

        self.train_device = torch.device(train_device)
        self.inference_device = torch.device(inference_device)
        self.verbose = verbose

        if binning_mode not in ["linear", "hist"]:
            raise ValueError("Only avaliable 'linear', 'hist' binning mods, but {} given".format(binning_mode))
        self.binning_mode = binning_mode
        self.bins_number = bins_number
        self.k = n_important
        self.learning_rate = learning_rate
        if n_epochs <= 0:
            raise ValueError("Epochs number should be positive, but {} given".format(n_epochs))
        self.n_epochs = n_epochs

        if not issubclass(optimizer, torch.optim.Optimizer):
            raise TypeError("Not torch.optim.Optimizer like optimizer format, {} given".format(type(optimizer)))
        self.optimizer = optimizer
        optimizer_params = optimizer_params or {}
        self.optim_params = {**optimizer_params, "lr": self.learning_rate}
        self.train_batch_size = train_batch_size
        self.valid_batch_size = valid_batch_size

        if temperature <= 0:
            raise ValueError("Temperature should be positive, but {} given".format(temperature))
        self.T = temperature

        if temp_anneal_factor <= 0:
            raise ValueError("Temperature annealing factor should be positive, but {} given".format(temp_anneal_factor))
        self.temp_anneal_factor = temp_anneal_factor

        if conv_filters <= 0:
            raise ValueError(
                "Number of filters in convolution layers should be positive, but {} given".format(conv_filters)
            )
        self.conv_filters = conv_filters
        if conv_ksize <= 0:
            raise ValueError(
                "Kernel size of filters in convolution layers should be positive, but {} given".format(conv_ksize)
            )
        self.conv_ksize = conv_ksize
        if hidden_dim <= 0:
            raise ValueError("Dimention of hidden layer should be positive, but {} given".format(hidden_dim))
        self.hidden_dim = hidden_dim
        if drop_rate >= 1 or drop_rate < 0:
            raise ValueError("Dropout rate should be in [0, 1), but {} given".format(drop_rate))
        self.drop_rate = drop_rate

        if importance_sampler not in ["gumbeltopk", "softsub"]:
            raise ValueError(
                "Only possible values for importance samper are 'gumbeltopk',"
                "'softsub', but {} given".format(importance_sampler)
            )
        self.importance_sampler = importance_sampler

        if embedder is None:
            self.embedder = None
            if embedding_dim is None:
                raise ValueError("At least embedding_dim or embedder should be not none")
            self.embedding_dim = embedding_dim
        elif isinstance(embedder, str):
            try:
                self.embedder = gensim.models.FastText.load(embedder).wv
            except:
                try:
                    self.embedder = gensim.models.FastText.load_fasttext_format(embedder).wv
                except:
                    self.embedder = gensim.models.KeyedVectors.load(embedder).wv
            self.embedding_dim = self.embedder.vector_size
        elif isinstance(embedder, dict):
            self.embedder = embedder
            self.embedding_dim = next(iter(embedder.values())).shape[0]
        elif isinstance(embedder, gensim.models.KeyedVectors):
            self.embedder = embedder
            self.embedding_dim = self.embedder.vector_size
        else:
            raise TypeError("Unknown embedder type: {}".format(embedder))
        self.trainable_embeds = trainable_embeds

        if not isinstance(max_vocab_length, int):
            raise TypeError("max_vocab_length should be int, but {} given".format(type(max_vocab_length)))
        elif max_vocab_length < -1 or max_vocab_length == 0:
            raise ValueError(
                "Only avaliable values for max_vocab_length: -1 or grater 0, but {} given".format(max_vocab_length)
            )
        self.max_vocab_length = max_vocab_length

        if gamma < 0:
            logger.info2("For now sparse token highlighting will be encouraged, since gamma < 0")
        self.gamma = gamma
        if gamma_anneal_factor <= 0:
            raise ValueError("Gamma annealing factor should be positive, but {} given".format(temp_anneal_factor))
        self.gamma_anneal_factor = gamma_anneal_factor
        self.explainers = {}
        self.random_seed = random_seed
        self.deterministic = deterministic
        seed_everything(random_seed, deterministic)

        if extreme_patience == 0:
            extreme_patience = n_epochs
        if patience == 0:
            patience = extreme_patience
        if patience > extreme_patience:
            logger.info2(
                "extreme_patience (={}) must be greater or equal patience (={}), now extreme_patience also ={}".format(
                    extreme_patience, patience, patience
                )
            )
            self.extreme_patience = patience
        self.patience = patience
        self.extreme_patience = extreme_patience

        if cache_dir is None:
            cache_dir = automl.autonlp_params["cache_dir"] or "./l2x_cache"
        if not isinstance(cache_dir, str):
            raise TypeError("Unknown type for cache_dir: {}".format(type(cache_dir)))
        os.makedirs(cache_dir, exist_ok=True)
        self.cache_dir = cache_dir
        self._checkpoint_path = cache_dir + "/l2x_checkpoint.pt"

    @property
    def n_important(self):
        return self.k

    @property
    def temperature(self):
        return self.T

    def fit(
        self,
        train_data: pd.DataFrame,
        valid_data: Optional[pd.DataFrame] = None,
        cols_to_explain: Optional[Union[str, List[str]]] = None,
    ):
        """Fit model for all columns in cols_to_explain.

        Target for L2X is the predictions of AutoML model. So, that you
        can pass any dataset there, but it recommend using training
        or validation datasets for the initial AutoML model.

        Args:
            train_data: Dataset, used for training the L2X model.
            valid_data: Validation dataset.
            cols_to_explain: Columns, for which the models will be built.
                If none the L2X models will be built corresponding to each
                textual column.

        """
        cols = self._get_cols(cols_to_explain)
        train_preds = self.automl.predict(train_data).data
        valid_preds = None
        if valid_data is not None:
            valid_preds = self.automl.predict(valid_data).data
        for col in cols:
            self.explainers[col] = self._fit_one(col, train_data, train_preds, valid_data, valid_preds)

    def _get_cols(self, cols_to_explain: Union[None, str, List[str]]) -> List[str]:
        """Handler for column names.

        Args:
            cols_to_explain: Explaining columns.

        Returns:
            Handled column names.

        """
        if isinstance(cols_to_explain, str):
            cols_to_explain = [cols_to_explain]
        if cols_to_explain is None:
            cols_to_explain = self.reader.cols_by_type("Text")
        else:
            for col in cols_to_explain:
                if self.roles[col].name != "Text":
                    raise ValueError("Column {} is not text column".format(col))

        return cols_to_explain

    def _fit_one(
        self,
        col_to_explain: str,
        train_data: pd.DataFrame,
        train_preds: np.ndarray,
        valid_data: Union[None, pd.DataFrame],
        valid_preds: Union[None, np.ndarray],
    ) -> "_L2XExplainer":
        """

        Args:
            col_to_explain: Explaining column.
            train_data: Dataset for training.
            train_preds: AutoML's (or any different model) predictions
                for train dataset.
            valid_data: Dataset for validation.
            valid_preds: AutoML's (or any different model) predictions
                for validation dataset.

        Returns:
            Explanation object.

        """
        train_tokenized = get_tokenized(train_data, col_to_explain, self.tokenizer)
        word_to_id, id_to_word = get_vocab(train_tokenized, self.max_vocab_length)
        word_to_id = WrappedVocabulary(word_to_id)
        weights_matrix = get_embedding_matrix(id_to_word, self.embedder, self.embedding_dim)
        train_data = map_tokenized_to_id(train_tokenized, word_to_id, self.k)
        train_dataset = get_len_dataset(train_data, train_preds)
        data_lens = self._get_lens(train_dataset)
        boundaries = self.calc_bins(data_lens, train_preds)
        train_dataloader = get_len_dataloader(train_dataset, self.train_batch_size, boundaries, mode="train")
        valid_dataloader = None
        if valid_data is not None:
            valid_tokenized = get_tokenized(valid_data, col_to_explain, self.tokenizer)
            valid_data = map_tokenized_to_id(valid_tokenized, word_to_id, self.k)
            valid_dataset = get_len_dataset(valid_data, valid_preds)
            valid_dataloader = get_len_dataloader(valid_dataset, self.valid_batch_size, boundaries, mode="valid")

        model = L2XModel(
            task_name=self.task_name,
            n_outs=self.n_outs,
            voc_size=len(word_to_id),
            embed_dim=self.embedding_dim,
            conv_filters=self.conv_filters,
            conv_ksize=self.conv_ksize,
            drop_rate=self.drop_rate,
            hidden_dim=self.hidden_dim,
            T=self.temperature,
            k=self.n_important,
            weights_matrix=weights_matrix,
            trainable_embeds=self.trainable_embeds,
            sampler=self.importance_sampler,
            anneal_factor=self.temp_anneal_factor,
        )
        train_loss_logs = []
        if valid_data is not None:
            valid_loss_logs = []
        self.train(model, train_dataloader, train_loss_logs, valid_dataloader, valid_loss_logs)

        return _L2XExplainer(
            model,
            col_to_explain,
            boundaries,
            self.tokenizer,
            word_to_id,
            id_to_word,
            self.inference_device,
            self.k,
            self.task_name,
            train_loss_logs,
            valid_loss_logs,
        )

    def train(
        self,
        model: L2XModel,
        train_dataloader: torch.utils.data.DataLoader,
        train_loss_logs: List[float],
        valid_dataloader: Optional[torch.utils.data.DataLoader] = None,
        valid_loss_logs: List[Union[float, None]] = None,
    ):
        """Trainer for L2X.

        Args:
            model: Model to train.
            train_dataloader: Dataloader, used for training.
            train_loss_logs: Train process logs.
                Contains train loss list.
            valid_dataloader: Dataloader, used for validation.
            valid_loss_logs: Train process logs.
                Contains train loss list.

        """
        loss = self._loss
        optimizer = self.optimizer(model.parameters(), **self.optim_params)
        model.to(self.train_device)
        prev_best = -1
        best_loss = torch.finfo(float).max
        gamma = self.gamma

        scheduler = ReduceLROnPlateau(optimizer, patience=self.patience)
        for epoch in range(self.n_epochs):
            train_loss = self._train_epoch(model, train_dataloader, loss, optimizer, self.train_device, gamma)
            valid_loss = self._validate(model, valid_dataloader, loss, self.train_device)
            train_loss_logs.append(train_loss)
            if valid_loss_logs is not None:
                valid_loss_logs.append(valid_loss)
            if self.verbose:
                if valid_loss is None:
                    logger.info3("Epoch: {}/{}, train loss: {}".format(epoch + 1, self.n_epochs, train_loss))
                else:
                    logger.info3(
                        "Epoch: {}/{}, train loss: {}, valid loss: {}".format(
                            epoch + 1, self.n_epochs, train_loss, valid_loss
                        )
                    )
            if valid_loss is not None:
                scheduler.step(valid_loss)
                if self.extreme_patience > 0 and valid_loss < best_loss:
                    best_loss = valid_loss
                    prev_best = epoch
                    torch.save(model.state_dict(), self._checkpoint_path)

                elif self.extreme_patience > 0 and epoch - prev_best > self.extreme_patience:
                    model.load_state_dict(torch.load(self._checkpoint_path))
                    break
            if epoch != self.n_epochs - 1:
                model.anneal()
                gamma *= self.gamma_anneal_factor

        model.cpu()

    def _train_epoch(
        self,
        model: L2XModel,
        train_dataloader: torch.utils.data.DataLoader,
        criterion: Union[torch.nn.modules.loss._Loss, Callable],
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        gamma: float,
    ) -> float:
        """Train only one epoch.

        Args:
            model: Trained L2X model.
            train_dataloader: Dataloader, used for training.
            criterion: Torch loss object. Unnormalized (biased) negative log-likelihood.
            optimizer: Optimizer that should be one or nothing.
            device: Device.
            gamma: Gamma.

        Returns:
            Accumalated loss.

        """
        model.train()
        if self.verbose:
            train_dataloader = tqdm(train_dataloader, desc="train", disable=False)

        accum_loss = 0.0
        iters = 0
        for data in train_dataloader:
            x = data["text"]
            x = x.to(device)
            y = data["target"]
            y = y.to(device)
            optimizer.zero_grad()
            pred, corr_pred = model(x)
            # Negative loglikelihood up to a constant
            nll_loss = criterion(pred, y)
            # Encouragement of neighbour tokens
            corr_loss = torch.mean(((corr_pred[:, 1:]) ** 2 * (corr_pred[:, :-1]) ** 2).sum(-1))
            # Not sure that optima of this pair of losses is the same
            # but dunno how get best validation score
            loss = nll_loss - gamma * corr_loss
            loss.backward()
            optimizer.step(),
            nll_loss = nll_loss.data.cpu().detach().numpy()
            accum_loss += nll_loss
            iters += 1

            if self.verbose:
                train_dataloader.set_description("train nll (loss={:.4f})".format(accum_loss / iters))

        return accum_loss / iters

    def _validate(
        self,
        model: L2XModel,
        valid_dataloader: Union[None, torch.utils.data.DataLoader],
        criterion: Union[torch.nn.modules.loss._Loss, Callable],
        device: torch.device,
    ) -> float:
        if valid_dataloader is None:
            return
        model.eval()
        model.to(device)
        loss = 0.0
        iters = 0
        with torch.no_grad():
            for data in valid_dataloader:
                x = data["text"]
                x = x.to(device)
                y = data["target"]
                y = y.to(device)
                pred, _ = model(x)
                nll_loss = criterion(pred, y).data.cpu().detach().numpy()
                loss += nll_loss
                iters += 1

        return loss / iters

    @staticmethod
    def _get_lens(dataset) -> np.ndarray:
        lens = []
        for inst in dataset:
            lens.append(inst["len"])
        return np.array(lens)

    def calc_bins(self, lens, target) -> np.ndarray:
        if self.binning_mode == "linear":
            return self._linear_bins(lens)
        elif self.binning_mode == "hist":
            return self._hist_bins(lens)
        elif self.bin_mode == "mi":
            return self._mi_bins(lens, target)

    def _linear_bins(self, lens) -> np.ndarray:
        return np.r_[0, np.linspace(lens.min(), lens.max() + 1, self.bins_number + 1)[1:]]

    def _hist_bins(self, lens) -> np.ndarray:
        return np.r_[0, np.hist(lens, self.bins_number)[1]]

    def _mi_bins(self, data, target) -> np.ndarray:
        raise NotImplementedError("Mutual information binning is not avaliable")

    def __getitem__(self, col) -> np.ndarray:
        if col not in self.explainers:
            raise ValueError("There is no explanation for {}".format(col))
        return self.explainers[col]


class _L2XExplainer:
    def __init__(
        self,
        model: L2XModel,
        col_to_explain: str,
        bins: List[int],
        tokenizer: Callable,
        word_to_id: Dict[str, int],
        id_to_word: Dict[int, str],
        inference_device: torch.device,
        n_important: int,
        task_name: str,
        train_loss_logs: List[float],
        valid_loss_logs: List[Union[float, None]],
    ):
        """
        Args:
            model: L2X explanation model.
            col_to_explain: Explaining column.
            bins: Binning used for training.
            tokenizer: Tokenizer function.
            word_to_id: Dictionary word to token-id.
            id_to_word: Dictionary token-id to word.
            inference_device: Inference device.
            n_important: Number of important tokens.
            task_name: Task name.
            train_loss_logs: Train process logs.
                Contains train loss list.
            valid_loss_logs: Valid process logs.
                Contains valid loss list.

        """
        self.model = model
        self.col_to_explain = col_to_explain
        self.bins = bins
        self.tokenizer = tokenizer
        self.word_to_id = word_to_id
        self.id_to_word = id_to_word
        self.inference_device = inference_device
        self.k = n_important
        self.task_name = task_name
        self.train_loss_logs = train_loss_logs
        self.valid_loss_logs = valid_loss_logs

    @property
    def train_loss(self):
        return self.train_loss_logs

    @property
    def valid_loss(self):
        return self.valid_loss_logs

    @property
    def n_train_epochs(self):
        return len(self.train_loss_logs)

    @property
    def n_important(self):
        return self.k

    def explain_instances(self, data: pd.DataFrame, batch_size: int = 1) -> "L2XExplanationsContainer":
        """Get explanations for data.

        Args:
            data: Data to explain.
            batch_size: Size of batch. Carefully, if batch_size > 0
                every document will be padded to max size of documents in abstract.

        Returns:
            Explaner.

        """
        data_tokenized = get_tokenized(data, self.col_to_explain, self.tokenizer)
        data = map_tokenized_to_id(data_tokenized, self.word_to_id, self.k)
        dataset = get_len_dataset(data, np.ones((len(data), 1)))

        dataloader = get_len_dataloader(dataset, batch_size, mode="test")
        masks = self._get_masks(dataloader)  # care the order.
        tokens = [self._tokens_to_text(dataset[i]["text"]) for i in range(len(dataset))]

        return L2XExplanationsContainer(tokens, masks, self.task_name)

    def _tokens_to_text(self, tokens):
        return list(map(self.id_to_word.get, tokens.numpy()))

    def _get_masks(self, dataloader: torch.utils.data.DataLoader) -> List[List[Number]]:
        important_tokens = []
        self.model.eval()
        self.model.to(self.inference_device)
        with torch.no_grad():
            for data in dataloader:
                x = data["text"]
                x = x.to(self.inference_device)
                _, imp_mask = self.model(x)
                important_tokens.extend([x.tolist() for x in imp_mask.cpu().numpy()])
            return important_tokens


class L2XExplanation:
    """
    Class for visualizing important tokens for single document.
    """

    _hightliting_color: str = "#6DB5F5"
    _background_color: str = "#FFFFFF"

    def __init__(self, tokens: List[str], mask: List[Number], task_name: str):
        """
        Args:
            tokens: Tokens of tokenized document.
            mask: Mask for important tokens. Non zero elements
                correspond to importatnt tokens.
            task_name: Task name.

        """
        if len(tokens) != len(mask):
            raise ValueError("Dimention mismatch for tokens and mask")
        self.tokens = tokens
        self.mask = mask
        self.task_name = task_name

    def visualize_in_notebook(self):
        """
        Visualization of important tokens in notebook.
        """
        from IPython.display import HTML
        from IPython.display import display_html

        token_weights = [(escape(x), y) for x, y in zip(self.tokens, self.mask)]
        html_code = draw_html(token_weights, self.task_name, self._hightliting_color, grad_line=False)
        display_html(HTML(html_code))


class L2XExplanationsContainer:
    """Container of explanations."""

    def __init__(self, docs: List[List[str]], masks: List[List[float]], task_name: str):
        """

        Args:
            docs: Tokenized documents.
            masks: Mask for importances.
            task_name: Task name.

        """
        self.docs = docs
        self.masks = masks
        self.task_name = task_name

    def __len__(self):
        return len(self.docs)

    def __getitem__(self, i):
        return L2XExplanation(self.docs[i], self.masks[i], self.task_name)

    def get_all(self):
        return [(x, y) for x, y in zip(self.docs, self.masks)]



"""Uplift meta-models."""
import copy
import logging

from abc import ABCMeta
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np

from pandas import DataFrame

from lightautoml.automl.base import AutoML
from lightautoml.automl.presets.tabular_presets import TabularAutoML
from lightautoml.tasks import Task
from lightautoml.utils.timer import Timer
from lightautoml.validation.np_iterators import UpliftIterator

from .utils import _get_target_role
from .utils import _get_treatment_role
from .utils import create_linear_automl


logger = logging.getLogger(__name__)


class NotTrainedError(Exception):
    """Exception: not trained model."""

    pass


class MetaLearner(metaclass=ABCMeta):
    """Base class for uplift meta-learner."""

    def __init__(
        self,
        base_task: Task,
        timeout: Optional[int] = None,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
    ):
        self.base_task = base_task
        self.timeout = timeout
        self.cpu_limit = cpu_limit
        self.gpu_ids = gpu_ids
        self._is_fitted = False

        self._timer = Timer()
        if timeout is not None:
            self._timer._timeout = timeout

    def fit(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit metalearner.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        """
        self._timer.start()
        self._fit(train_data, roles, verbose)
        self._is_fitted = True
        if self._timer.time_limit_exceeded():
            logger.warning(
                "{} is trained, but time limit exceeded.",
                self.__class__.__name__,
            )

    def predict(self, data: DataFrame) -> Tuple[np.ndarray, ...]:
        """Predict treatment effects.

        Args:
            data: Dataset to perform inference.

        Returns:
            treatment_effect: Predictions of treatment effects
            effect_w_interaction: Predictions of base task values on treated-group
            effect_wo_interaction: Predictions of base task values on control-group

        """
        if not self._is_fitted:
            raise NotTrainedError()
        return self._predict(data)

    @abstractmethod
    def _fit(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        pass

    @abstractmethod
    def _predict(self, data: DataFrame):
        pass

    def _get_default_learner(self, task: Task):
        return create_linear_automl(task)

    def _get_task(self, learner: AutoML) -> Task:
        if isinstance(learner, TabularAutoML):
            return learner.task
        elif isinstance(learner, AutoML):
            return learner.reader.task
        else:
            raise RuntimeError("Can't extract 'task' from learner")

    def _check_timer(self):
        if self._timer.time_limit_exceeded():
            logger.warning(
                "MetaLearner '%s' isn't trained, because time limit was exceeded",
                self.__class__.__name__,
            )
            raise NotTrainedError()


class SLearner(MetaLearner):
    """SLearner.

    `SLearner` - is an 'meta' model using 'treatment' column as a feature.

    """

    def __init__(
        self,
        learner: Optional[AutoML] = None,
        base_task: Optional[Task] = None,
        timeout: Optional[int] = None,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
    ):
        if base_task is None:
            if learner is not None:
                base_task = self._get_task(learner)
            else:
                raise RuntimeError('Must specify any of learners or "base_task"')

        super().__init__(base_task, timeout, cpu_limit, gpu_ids)

        if learner is None:
            self.learner = self._get_default_learner(base_task)
        else:
            self.learner = learner

        self._treatment_col: str

    def _fit(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit metalearner.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        """
        (
            treatment_role,
            treatment_col,
        ) = _get_treatment_role(roles)
        self._treatment_col = treatment_col

        uplift_roles = copy.deepcopy(roles)
        uplift_roles.pop(treatment_role)

        self.learner.fit_predict(
            train_data,
            uplift_roles,
            verbose=verbose,
        )

    def _predict(self, data: DataFrame) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predict treatment effects.

        Args:
            data: Dataset to perform inference.

        Returns:
            treatment_effect: Predictions of treatment effects
            effect_w_interaction: Predictions of base task values on treated-group
            effect_wo_interaction: Predictions of base task values on control-group

        """
        data_c = data.copy()

        data_c[self._treatment_col] = 0
        control_pred = self.learner.predict(data_c).data.ravel()

        data_c[self._treatment_col] = 1
        treatment_pred = self.learner.predict(data_c).data.ravel()

        uplift_pred = treatment_pred - control_pred

        return (
            uplift_pred,
            treatment_pred,
            control_pred,
        )


class TLearner(MetaLearner):
    """TLearner.

    `TLearner` - is an 'meta' model which uses a two separated models.

    Each model is trained on it's own group (treatment/control).

    The 'meta' model prediction is a subtraction predictions of 'treatment' model and 'control' model.


    Args:
        treatment_learner: AutoML model, if `None` then will be used model by default
        control_learner: AutoML model, if `None` then will be used model by default
        base_task: task
        timeout: Timeout
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.

    """

    def __init__(
        self,
        treatment_learner: Optional[AutoML] = None,
        control_learner: Optional[AutoML] = None,
        base_task: Optional[Task] = None,
        timeout: Optional[int] = None,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
    ):
        assert any(
            x is not None
            for x in [
                treatment_learner,
                control_learner,
                base_task,
            ]
        ), 'Must specify any of learners or "base_task"'

        if base_task is None:
            if treatment_learner is not None:
                base_task = self._get_task(treatment_learner)
            elif control_learner is not None:
                base_task = self._get_task(control_learner)

        super().__init__(base_task, timeout, cpu_limit, gpu_ids)

        self.treatment_learner = (
            treatment_learner if treatment_learner is not None else self._get_default_learner(self.base_task)
        )
        self.control_learner = (
            control_learner if control_learner is not None else self._get_default_learner(self.base_task)
        )

    def _fit(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit metalearner.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        """
        (
            treatment_role,
            treatment_col,
        ) = _get_treatment_role(roles)

        new_roles = copy.deepcopy(roles)
        new_roles.pop(treatment_role)

        control_train_data = train_data[train_data[treatment_col] == 0]
        treatment_train_data = train_data[train_data[treatment_col] == 1]

        control_train_data.drop(treatment_col, axis=1, inplace=True)
        treatment_train_data.drop(treatment_col, axis=1, inplace=True)

        self.treatment_learner.fit_predict(
            treatment_train_data,
            new_roles,
            verbose=verbose,
        )
        self._check_timer()
        self.control_learner.fit_predict(
            control_train_data,
            new_roles,
            verbose=verbose,
        )

    def _predict(self, data: Any) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predict treatment effects.

        Args:
            data: Dataset to perform inference.

        Returns:
            treatment_effect: Predictions of treatment effects
            effect_w_interaction: Predictions of base task values on treated-group
            effect_wo_interaction: Predictions of base task values on control-group

        """
        treatment_pred = self.treatment_learner.predict(data).data.ravel()
        control_pred = self.control_learner.predict(data).data.ravel()

        uplift = treatment_pred - control_pred

        return (
            uplift,
            treatment_pred,
            control_pred,
        )


class T2Learner(MetaLearner):
    """T2Learner.

    `T2Learner` - is a modification of `TLearner` model.

    Difference of original approach is a training scheme (`UpliftIterator`).

    To train the base task of the model (treatment/control), use both parts of datasets (treatment/control) with
    additional feature `treatment_flag`.

    Each model is tuned on corresponding dataset.


    Args:
        treatment_learner: AutoML model, if `None` then will be used model by default
        control_learner: AutoML model, if `None` then will be used model by default
        n_uplift_iterator_folds: Number of folds
        base_task: task
        timeout: Timeout
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.

    """

    def __init__(
        self,
        treatment_learner: Optional[AutoML] = None,
        control_learner: Optional[AutoML] = None,
        n_uplift_iterator_folds: int = 5,
        base_task: Optional[Task] = None,
        timeout: Optional[int] = None,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
    ):
        if base_task is None:
            if treatment_learner is not None:
                base_task = self._get_task(treatment_learner)
            elif control_learner is not None:
                base_task = self._get_task(control_learner)
            else:
                raise RuntimeError('Must specify any of learners or "base_task"')

        super().__init__(base_task, timeout, cpu_limit, gpu_ids)

        self._n_uplift_iterator_folds = n_uplift_iterator_folds

        self.treatment_learner = (
            treatment_learner if treatment_learner is not None else self._get_default_learner(self.base_task)
        )
        self.control_learner = (
            control_learner if control_learner is not None else self._get_default_learner(self.base_task)
        )

    def _fit(self, train_data: DataFrame, roles: Dict):
        """Fit metalearner.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles

        """
        (
            treatment_role,
            treatment_col,
        ) = _get_treatment_role(roles)
        _, target_col = _get_target_role(roles)
        self._treatment_col = treatment_col

        new_roles = copy.deepcopy(roles)
        new_roles.pop(treatment_role)

        train_data_c = train_data.copy()
        treatment_values = train_data_c[treatment_col].values
        target_values = train_data[target_col].values

        treatment_iterator = UpliftIterator(
            treatment_values,
            target_values,
            True,
            self.base_task,
            self._n_uplift_iterator_folds,
        )
        self.treatment_learner.fit_predict(
            train_data_c,
            new_roles,
            cv_iter=treatment_iterator,
        )

        control_iterator = UpliftIterator(
            treatment_values,
            target_values,
            False,
            self.base_task,
            self._n_uplift_iterator_folds,
        )
        self.control_learner.fit_predict(
            train_data_c,
            new_roles,
            cv_iter=control_iterator,
        )

    def _predict(self, data: DataFrame):
        """Predict treatment effects.

        Args:
            data: Dataset to perform inference.

        Returns:
            treatment_effect: Predictions of treatment effects
            effect_w_interaction: Predictions of base task values on treated-group
            effect_wo_interaction: Predictions of base task values on control-group

        """
        data_с = data.copy()
        data_с[self._treatment_col] = True
        treatment_pred = self.treatment_learner.predict(data_с).data.ravel()
        data_с[self._treatment_col] = False
        control_pred = self.control_learner.predict(data_с).data.ravel()

        uplift = treatment_pred - control_pred

        return (
            uplift,
            treatment_pred,
            control_pred,
        )


class TDLearner(MetaLearner):
    """TDLearner.

    `TDLearner` - is an 'meta' model which uses a two models (the one model depends on the prediction of another model).

    The 'meta' model prediction is a subtraction predictions of 'treatment' model and 'control' model.


    Args:
        treatment_learner: AutoML model, if `None` then will be used model by default
        control_learner: AutoML model, if `None` then will be used model by default
        base_task: task
        timeout: Timeout
        dependent_group: Value := {0 , 1}. Dependent group on the prediction of another group,
            If `None` is dependent group will be a large group by size
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.

    """

    def __init__(
        self,
        treatment_learner: Optional[AutoML] = None,
        control_learner: Optional[AutoML] = None,
        base_task: Optional[Task] = None,
        timeout: Optional[int] = None,
        dependent_group: Optional[int] = None,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
    ):
        assert any(
            x is not None
            for x in [
                treatment_learner,
                control_learner,
                base_task,
            ]
        ), 'Must specify any of learners or "base_task"'

        if base_task is None and (treatment_learner is None or control_learner is None):
            if treatment_learner is not None:
                base_task = self._get_task(treatment_learner)
            elif control_learner is not None:
                base_task = self._get_task(control_learner)

        super().__init__(base_task, timeout, cpu_limit, gpu_ids)

        self.treatment_learner = (
            treatment_learner if treatment_learner is not None else self._get_default_learner(self.base_task)
        )
        self.control_learner = (
            control_learner if control_learner is not None else self._get_default_learner(self.base_task)
        )

        self._other_group_pred_col = "__OTHER_GROUP_PREDICTION__"
        self._dependent_group: Optional[int] = dependent_group

    def _fit(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit metalearner.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        """
        (
            treatment_role,
            treatment_col,
        ) = _get_treatment_role(roles)

        self._set_dependent_group(train_data[treatment_col].mean())

        new_roles = copy.deepcopy(roles)
        new_roles.pop(treatment_role)

        control_train_data = train_data[train_data[treatment_col] == 0]
        treatment_train_data = train_data[train_data[treatment_col] == 1]

        control_train_data.drop(treatment_col, axis=1, inplace=True)
        treatment_train_data.drop(treatment_col, axis=1, inplace=True)

        if self._dependent_group == 1:
            dependent_train_data = treatment_train_data
            dependent_learner = self.treatment_learner

            independent_train_data = control_train_data
            independent_learner = self.control_learner
        else:
            dependent_train_data = control_train_data
            dependent_learner = self.control_learner

            independent_train_data = treatment_train_data
            independent_learner = self.treatment_learner

        independent_learner.fit_predict(independent_train_data, new_roles)
        self._check_timer()
        sg_oof_pred = independent_learner.predict(dependent_train_data).data.ravel()
        dependent_train_data[self._other_group_pred_col] = sg_oof_pred
        self._check_timer()
        dependent_learner.fit_predict(dependent_train_data, new_roles)

    def _predict(self, data: Any) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predict treatment effects.

        Args:
            data: Dataset to perform inference.

        Returns:
            treatment_effect: Predictions of treatment effects
            effect_w_interaction: Predictions of base task values on treated-group
            effect_wo_interaction: Predictions of base task values on control-group

        """
        data_c = data.copy()

        if self._dependent_group == 1:
            (dependent_learner, independent_learner,) = (
                self.treatment_learner,
                self.control_learner,
            )
        else:
            (dependent_learner, independent_learner,) = (
                self.control_learner,
                self.treatment_learner,
            )

        independent_pred = independent_learner.predict(data_c).data.ravel()
        data_c[self._other_group_pred_col] = independent_pred
        dependent_pred = dependent_learner.predict(data_c).data.ravel()

        if self._dependent_group == 1:
            control_pred, treatment_pred = (
                independent_pred,
                dependent_pred,
            )
        else:
            control_pred, treatment_pred = (
                dependent_pred,
                independent_pred,
            )

        uplift = treatment_pred - control_pred

        return (
            uplift,
            treatment_pred,
            control_pred,
        )

    def _set_dependent_group(self, treatment_rate: float):
        if self._dependent_group is None:
            self._dependent_group = 1 if treatment_rate > 0.5 else 0


class XLearner(MetaLearner):
    """XLearner.

    XLearner - is a 'meta' model which use approach from `TLearner` `meta` model.

    The learning algorithm:

    Step #1 (Propensity score): Train the model to distinguish between the target and control groups.
    Step #2 (Outcome): Train two models on treatment/control group to predict base task, named this `outcome` model.
    Step #3 (Effect): Train models to predict difference between true outcome of treatment part dataset and prediction of
        `outcome-control` model (step 1) on treatment group dataset, and same for outcome, but with negative sign,
        named this `effect` model.

    Final prediction of `XLearner` is weighted sum of `effect` models (treatment/control), where weights is propensity score.


    Args:
        outcome_learners: Models predict `outcome` (base task) for each group (treatment/control),
            base task can be classification or regression task.
            It can be: two models, one model or nothing.
            If there is one model, then it will used for both groups.
            If `None` then will be used model by default.
        effect_learners:  Models predict treatment effect. (task must be 'reg')
            It can be: two models, one model or nothing.
            If there is one model, then it will used for both groups.
            If `None` then will be used model by default.
        propensity_learner: Model predicts treatment group membership,
            If `None` then will be used model by default
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.
        base_task: Task - 'binary' or 'reg'

    """

    def __init__(
        self,
        outcome_learners: Optional[Sequence[AutoML]] = None,
        effect_learners: Optional[Sequence[AutoML]] = None,
        propensity_learner: Optional[AutoML] = None,
        base_task: Optional[Task] = None,
        timeout: Optional[int] = None,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
    ):
        if (outcome_learners is None or len(outcome_learners) == 0) and base_task is None:
            raise RuntimeError('Must specify any of learners or "base_task"')

        if outcome_learners is not None and len(outcome_learners) > 0:
            base_task = self._get_task(outcome_learners[0])
            super().__init__(self._get_task(outcome_learners[0]))

        super().__init__(base_task, timeout, cpu_limit, gpu_ids)

        self.learners: Dict[str, Union[Dict[str, AutoML], AutoML]] = {
            "outcome": {},
            "effect": {},
        }
        if propensity_learner is None:
            self.learners["propensity"] = self._get_default_learner(Task("binary"))
        else:
            self.learners["propensity"] = propensity_learner

        if outcome_learners is None or len(outcome_learners) == 0:
            self.learners["outcome"]["control"] = self._get_default_learner(self.base_task)
            self.learners["outcome"]["treatment"] = self._get_default_learner(self.base_task)
        elif len(outcome_learners) == 1:
            self.learners["outcome"]["control"] = outcome_learners[0]
            self.learners["outcome"]["treatment"] = copy.deepcopy(outcome_learners[0])
        elif len(outcome_learners) == 2:
            self.learners["outcome"]["control"] = outcome_learners[0]
            self.learners["outcome"]["treatment"] = outcome_learners[1]
        else:
            raise RuntimeError('The number of "outcome_learners" must be 0/1/2')

        if effect_learners is None or len(effect_learners) == 0:
            self.learners["effect"]["control"] = self._get_default_learner(Task("reg"))
            self.learners["effect"]["treatment"] = self._get_default_learner(Task("reg"))
        elif len(effect_learners) == 1:
            self.learners["effect"]["control"] = effect_learners[0]
            self.learners["effect"]["treatment"] = copy.deepcopy(effect_learners[0])
        elif len(effect_learners) == 2:
            self.learners["effect"]["control"] = effect_learners[0]
            self.learners["effect"]["treatment"] = effect_learners[1]
        else:
            raise RuntimeError('The number of "effect_learners" must be 0/1/2')

    def _fit(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit metalearner.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        """
        self._fit_propensity_learner(train_data, roles, verbose)
        self._fit_outcome_learners(train_data, roles, verbose)
        self._fit_effect_learners(train_data, roles, verbose)

    def _fit_propensity_learner(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit propensity score.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        """
        propensity_roles = copy.deepcopy(roles)

        (
            target_role,
            target_col,
        ) = _get_target_role(roles)
        propensity_roles.pop(target_role)

        (
            treatment_role,
            treatment_col,
        ) = _get_treatment_role(roles)
        propensity_roles.pop(treatment_role)
        propensity_roles["target"] = treatment_col

        train_cp = train_data.copy()
        train_cp.drop(target_col, axis=1, inplace=True)

        self.learners["propensity"].fit_predict(
            train_cp,
            propensity_roles,
            verbose=verbose,
        )

    def _fit_outcome_learners(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit outcome.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        """
        (
            treatment_role,
            treatment_col,
        ) = _get_treatment_role(roles)
        outcome_roles = copy.deepcopy(roles)
        outcome_roles.pop(treatment_role)

        for (
            group_name,
            outcome_learner,
        ) in self.learners["outcome"].items():
            self._check_timer()
            group = 1 if group_name == "treatment" else 0

            train_data_outcome = train_data[train_data[treatment_col] == group].copy()
            train_data_outcome.drop(
                treatment_col,
                axis=1,
                inplace=True,
            )

            outcome_learner.fit_predict(
                train_data_outcome,
                outcome_roles,
                verbose=verbose,
            )

    def _fit_effect_learners(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit treatment effects.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        """
        (
            treatment_role,
            treatment_col,
        ) = _get_treatment_role(roles)
        _, target_col = _get_target_role(roles)

        effect_roles: Dict = copy.deepcopy(roles)
        effect_roles.pop(treatment_role)

        for (
            group_name,
            effect_learner,
        ) in self.learners["effect"].items():
            self._check_timer()
            group = 1 if group_name == "treatment" else 0
            opposite_group_name = "treatment" if group_name == "control" else "control"

            train_data_effect = train_data[train_data[treatment_col] == group].copy()
            train_data_effect.drop(
                treatment_col,
                axis=1,
                inplace=True,
            )

            outcome_pred = self.learners["outcome"][opposite_group_name].predict(train_data_effect).data.ravel()
            train_data_effect[target_col] = train_data_effect[target_col] - outcome_pred

            if group_name == "control":
                train_data_effect[target_col] *= -1

            train_data_effect = train_data_effect[train_data_effect[target_col].notnull()]

            effect_learner.fit_predict(
                train_data_effect,
                effect_roles,
                verbose=verbose,
            )

    def _predict(self, data: Any) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predict treatment effects.

        Args:
            data: Dataset to perform inference.

        Returns:
            treatment_effect: Predictions of treatment effects
            effect_w_interaction: Predictions of base task values on treated-group
            effect_wo_interaction: Predictions of base task values on control-group

        """
        outcome_control_pred = self.learners["outcome"]["control"].predict(data).data.ravel()
        outcome_treatment_pred = self.learners["outcome"]["treatment"].predict(data).data.ravel()

        propensity_score = self.learners["propensity"].predict(data).data.ravel()
        uplift_control_pred = self.learners["effect"]["control"].predict(data).data.ravel()
        uplift_treatment_pred = self.learners["effect"]["treatment"].predict(data).data.ravel()
        uplift = propensity_score * uplift_treatment_pred + (1.0 - propensity_score) * uplift_control_pred

        return (
            uplift,
            outcome_treatment_pred,
            outcome_control_pred,
        )


class RLearner(MetaLearner):
    r"""RLearner.

    m(x) - the conditional mean outcome
    e(x) - the propensity score
    tau(x) - the treatment effect

    .. math::
        \tau(\cdot) = argmin_{\tau} \sum_{i} \Big[ (Y_i - m(X_i)) - (W_i - e(X_i))\tau(X_i) \Big]^2


    Args:
        propensity_learner: AutoML model, if `None` then will be used model by default (task must be 'binary')
        mean_outcome_learner: AutoML model, if `None` then will be used model by default
        effect_learner: AutoML model, if `None` then will be used model by default (task must be 'reg')
        base_task: task
        timeout: Timeout
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.

    """

    _epsi = 10 ** -5

    def __init__(
        self,
        propensity_learner: Optional[AutoML] = None,
        mean_outcome_learner: Optional[AutoML] = None,
        effect_learner: Optional[AutoML] = None,
        base_task: Optional[Task] = None,
        timeout: Optional[int] = None,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
    ):
        if propensity_learner is not None and self._get_task(propensity_learner).name != "binary":
            raise RuntimeError("Task of 'propensity_learner' must be 'binary'")

        if mean_outcome_learner is None and base_task is None:
            raise RuntimeError("Must specify 'mean_outcome_learner' or base_task")

        if effect_learner is not None and self._get_task(effect_learner).name != "reg":
            raise RuntimeError("Task of effect_learner must be 'reg'")

        super().__init__(base_task, timeout, cpu_limit, gpu_ids)

        self.propensity_learner: AutoML
        self.mean_outcome_learner: AutoML
        self.effect_learner: AutoML

        no_learners = (propensity_learner is None) and (mean_outcome_learner is None) and (effect_learner is None)
        tabular_timeout = timeout / 3 if no_learners and timeout is not None else None

        if propensity_learner is None:
            self.propensity_learner = TabularAutoML(
                task=Task("binary"),
                timeout=tabular_timeout,
                reader_params={"treatment": {"force_input": True}},
            )
        else:
            self.propensity_learner = propensity_learner

        if mean_outcome_learner is not None:
            self.mean_outcome_learner = mean_outcome_learner
            self.base_task = self._get_task(mean_outcome_learner)
        elif base_task is not None:
            self.mean_outcome_learner = TabularAutoML(
                task=base_task,
                timeout=tabular_timeout,
                reader_params={"treatment": {"force_input": True}},
            )

        if effect_learner is None:
            self.effect_learner = TabularAutoML(
                task=Task("reg"),
                timeout=tabular_timeout,
                reader_params={"treatment": {"force_input": True}},
            )
        else:
            self.effect_learner = effect_learner

    def _fit(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit metalearner.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        """
        propensity_pred = self._fit_predict_propensity_learner(train_data, roles, verbose)
        self._check_timer()
        mean_outcome_pred = self._fit_predict_mean_outcome_learner(train_data, roles, verbose)
        self._check_timer()
        self._fit_effect_learner(
            train_data,
            roles,
            propensity_pred,
            mean_outcome_pred,
            verbose,
        )

    def _predict(self, data: Any) -> Tuple[np.ndarray, None, None]:
        """Predict treatment effects.

        Args:
            data: Dataset to perform inference.

        Returns:
            treatment_effect: Predictions of treatment effects
            None: Plug
            None: Plug

        """
        return (
            self.effect_learner.predict(data).data.ravel(),
            None,
            None,
        )

    def _fit_predict_propensity_learner(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit propensity score.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        Returns:
            oof propensity score prediction

        """
        propensity_roles = copy.deepcopy(roles)

        (
            target_role,
            target_col,
        ) = _get_target_role(roles)
        propensity_roles.pop(target_role)

        (
            treatment_role,
            treatment_col,
        ) = _get_treatment_role(roles)
        propensity_roles.pop(treatment_role)
        propensity_roles["target"] = treatment_col

        train_cp = train_data.copy()
        train_cp.drop(target_col, axis=1, inplace=True)

        propensity_pred = self.propensity_learner.fit_predict(
            train_cp,
            propensity_roles,
            verbose=verbose,
        ).data.ravel()

        return propensity_pred

    def _fit_predict_mean_outcome_learner(
        self,
        train_data: DataFrame,
        roles: Dict,
        verbose: int = 0,
    ):
        """Fit mean outcome.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            verbose: Verbose

        Returns:
            oof mean outcome prediction

        """
        outcome_roles = copy.deepcopy(roles)

        # target_role, target_col = _get_target_role(roles)

        (
            treatment_role,
            treatment_col,
        ) = _get_treatment_role(roles)
        outcome_roles.pop(treatment_role)

        train_cp = train_data.copy()
        train_cp.drop(treatment_col, axis=1, inplace=True)

        mean_outcome_pred = self.mean_outcome_learner.fit_predict(train_cp, outcome_roles).data.ravel()

        return mean_outcome_pred

    def _fit_effect_learner(
        self,
        train_data: DataFrame,
        roles: Dict,
        propensity_pred: np.ndarray,
        mean_outcome_pred: np.ndarray,
        verbose: int = 0,
    ):
        """Fit treatment effects.

        Args:
            train_data: Dataset to train
            roles: Roles dict with 'treatment' roles
            propensity_pred: oof-prediction of propensity_learner
            mean_outcome_pred: oof-prediction of mean_outcome_learner
            verbose: Verbose

        """
        effect_roles = copy.deepcopy(roles)

        _, target_col = _get_target_role(roles)
        train_target = train_data[target_col]

        (
            treatment_role,
            treatment_col,
        ) = _get_treatment_role(roles)
        train_treatment = train_data[treatment_col]
        effect_roles.pop(treatment_role)

        weights = train_treatment - propensity_pred + self._epsi

        train_cp = train_data.copy()
        train_cp.drop(treatment_col, axis=1, inplace=True)
        train_cp[target_col] = (train_target - mean_outcome_pred) / weights
        train_cp["__WEIGHTS__"] = weights ** 2
        effect_roles["weights"] = "__WEIGHTS__"

        train_cp = train_cp[train_cp[target_col].notnull()]

        self.effect_learner.fit_predict(
            train_cp,
            effect_roles,
            verbose=verbose,
        )
